/*****************************************************************************/
/* N.A.D.A.R. - A network tank battle game.                                  */
/*                                                                           */
/* N.A.D.A.R. Copyright (c) 2000-2001 Sakai Hiroaki.                         */
/* All Rights Reserved.                                                      */
/*                                                                           */
/* N.A.D.A.R. is a network tank battle game. You get into a tank,            */
/* operate it, and destroy other tanks to survive!                           */
/*****************************************************************************/

/*****************************************************************************/
/* 饤ȤΥᥤ                                                      */
/*****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <signal.h>

#include "Client.h"
#include "arguments.h"
#include "etc.h"

Client client = NULL;

static int quit2()
{
  Player player;
  unsigned short int command;
  unsigned short int length;
  int i;

  if (!client) return (1);

  player = ClientGetPlayer(client);
  PathSendNoneData(PlayerGetPath(player), DELETE_ME);
  PathFlushSendBuffer(PlayerGetPath(player));

  while (1) {
    for (i = 0; i < 30; i++) {
      if (PathCheckData(PlayerGetPath(player), &command, &length)) break;
      usleep (100000);
    }

    if (!PathCheckData(PlayerGetPath(player), &command, &length)) return (1);

    if (command == DELETE_YOU) {
      PathRecvNoneData(PlayerGetPath(player), &command);
      if (command != DELETE_YOU)
	ErrorExit("Unknown command.");
      break;
    }
    PathDiscardData(PlayerGetPath(player), &command);
  }
  return (0);
}

void quit3()
{
  if (client) client = ClientDestroy(client);
  fprintf(stderr, "Terminated.\n");
  exit (0);
}

static void quit(int value)
{
  if (client) quit2();
  quit3();
}

static char * player_names[] = {
  "Bolken", "WildDog", "Hornet", "Airi", "Akana",
  "GreenTea", "Dindon", "Ripple", "RingOnTheWater", "Blaky",
  "HalfPot", "SmartBomb", "Flogger", "DirtyHorse", "HungryStomach",
  "SlamPoint", "RattleSnake", "GrindBoy", "SqueakVoice", "JaggerEdge",
  "Belch", "WoodenDoll", "NimbleBoy", "PlayingHookey", "DrippingWet",
  "SillyBoy", "VeryBottom", "RapidlyLike", "Nawo", "ScrewDriver",
  "NezumiKozou", "Nababa", "Yuuki", "NipponoFace", "WhoAmI?",
  "FuurinKazan", "Polpfun", "LZ", "I.L.J.", "Folcman",
  "Goemon", "Jorjor", "BerkenIII", "Kame", "Sushi",
  "JarringSound", "Jumble", "MixJuice", "CupCake", "CottageCheese",
  "Gingin", "Kennel", "CoffeeBean", "JunkFood", "HighwayPatrol"
};

#define PLAYER_NAMES_NUMBER ((int)(sizeof(player_names)/sizeof(player_names[0])))

static void show_help()
{
  printf("\n");
  printf("                                 N.A.D.A.R.\n");
  printf("                        A network tank battle game.\n");
  printf("\n");
  printf("             N.A.D.A.R. Copyright (c) 2000-2001 Sakai Hiroaki.\n");
  printf("                            All Rights Reserved.\n");
  printf("\n");
  printf("This is ");
  printf(VERSION);
  printf(" client. Be careful that different versions of N.A.D.A.R. are unconnectable.\n");
  printf("\n");
  printf("N.A.D.A.R. is a client-server model program. ");
  printf("This is N.A.D.A.R.\'s client program.\n");
  printf("\n");
  printf("If you want to play N.A.D.A.R., You have to start server program\n");
  printf("on a server machine as\n");
  printf("\n");
  printf("  %% nadars\n");
  printf("\n");
  printf("And start client program as\n");
  printf("\n");
  printf("  %% nadar -s [Server\'s hostname] -p [Player\'s name]\n");
  printf("\n");
  printf("If playing speed is very slow, ");
  printf("run clients with size option as below.\n");
  printf("\n");
  printf("  %% nadar -s [Server\'s hostname] -p [Player\'s name] -size 30\n");
  printf("\n");
  printf("If the performance of client machine is very high, try big size.\n");
  printf("\n");
  printf("  %% nadar -s [Server\'s hostname] -p [Player\'s name] -size 120\n");
  printf("\n");
}

static void show_options()
{
  printf("Options :\n\n");
  printf("\t-display [displayname] : The name of the X server to use.\n");
  printf("\t-geometry [+X+Y] : The placement of the window on the screen.\n");
  printf("\t-name [name] : The name of the application.\n");
  printf("\t-bg, -background [color] : The color of background.\n");
  printf("\t-h, -help : Show help messages.\n");
  printf("\t-s, -server [server\'s hostname] : The hostname of the server.\n");
  printf("\t-p, -player [player\'s name] : The name of the player.\n");
  printf("\t-size [display size] : The size of the game screen by percent.\n");
  printf("\t-port : The port number of socket. Default is 12010.\n");
  printf("\t-direct-draw : Fast draw mode.\n");
  printf("\t-no-direct-draw : No flickering mode. (Default)\n");
  printf("\n");
}

static void help()
{
  show_help();
  show_options();
  exit (0);
}

static char * server_name = NULL;
static char * player_name = NULL;
static int port           = PORT;
static int size           = DISPLAY_DEFAULT_SIZE;
static int direct_draw    = DEFAULT_DIRECT_DRAW_FLAG;

static Argument arguments[] = {
  {"-h",                ARGUMENTS_FUNCTION, (void *)(&help)},
  {"-help",             ARGUMENTS_FUNCTION, (void *)(&help)},
  {"-s",                ARGUMENTS_STRING,   (void *)(&server_name)},
  {"-server",           ARGUMENTS_STRING,   (void *)(&server_name)},
  {"-p",                ARGUMENTS_STRING,   (void *)(&player_name)},
  {"-player",           ARGUMENTS_STRING,   (void *)(&player_name)},
  {"-port",             ARGUMENTS_INTEGER,  (void *)(&port)},
  {"-size",             ARGUMENTS_INTEGER,  (void *)(&size)},
  {"-direct-draw",      ARGUMENTS_FLAG_ON,  (void *)(&direct_draw)},
  {"-no-direct-draw",   ARGUMENTS_FLAG_OFF, (void *)(&direct_draw)},
  {NULL,                ARGUMENTS_NONE,     NULL}
};

int main(int argc, char * argv[])
{
  int i;

  signal(SIGINT , quit);
  signal(SIGTERM, quit);

  InitializeRandom();

  Arguments_Read(&argc, argv, arguments);

  if (server_name == NULL) {
    show_help();
    exit (1);
  }

  if (player_name == NULL)
    player_name = player_names[GetRandomValue(PLAYER_NAMES_NUMBER)];

  if (port < 1000) port = PORT;

  if (size < 3 || size > 500) size = DISPLAY_DEFAULT_SIZE;

  client = ClientCreate(server_name, port, player_name, size, direct_draw,
			&argc, argv);
  if (client == NULL) ErrorExit("Cannot create client.");

  for (i = 1; i < argc; i++) {
    fprintf(stderr, "Unknown option : %s\n", argv[i]);
  }

  ClientStart(client);

  client = ClientDestroy(client);

  exit (0);
}

/*****************************************************************************/
/* End of File.                                                              */
/*****************************************************************************/
