/*****************************************************************************/
/* N.A.D.A.R. - A network tank battle game.                                  */
/*                                                                           */
/* N.A.D.A.R. Copyright (c) 2000-2001 Sakai Hiroaki.                         */
/* All Rights Reserved.                                                      */
/*                                                                           */
/* N.A.D.A.R. is a network tank battle game. You get into a tank,            */
/* operate it, and destroy other tanks to survive!                           */
/*****************************************************************************/

/*****************************************************************************/
/* Game 饹(Υ饹)Υ饤֥                             */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                   */
/*****************************************************************************/

#include "GameP.h"

/*****************************************************************************/
/* ؿ                                                                */
/*****************************************************************************/

/*===========================================================================*/
/* ѿμ                                                    */
/*===========================================================================*/

Stage GameGetStage(Game game) { return (game->stage); }
ObjList GameGetPlayerList(Game game) { return (game->player_list); }
Path GameGetRequest(Game game) { return (game->request); }

/*===========================================================================*/
/* ४СȽ                                                      */
/*===========================================================================*/

static int GameIsOver(Game game)
{
  return (0);
}

/*===========================================================================*/
/* ֥ȤȺ                                                  */
/*===========================================================================*/

/*---------------------------------------------------------------------------*/
/* ֥Ȥ                                                        */
/*---------------------------------------------------------------------------*/

Game GameCreate(char * hostname, unsigned short int port,
		int map_width, int map_height, int width, int height,
		int peeping, int backward, int interval_time, int timeout_time,
		int max_players)
{
  Game game;

  game = (Game)malloc(sizeof(_Game));
  if (game == NULL) ErrorExit("Cannot allocate memory for game.");

  game->stage = NULL;

  game->player_list = ObjList_Create();
  if (game->player_list == NULL) ErrorExit("Cannot create game->player_list.");

  game->request = PathCreateToAcceptRequestsFromClients(hostname, port, 5);
  if (game->request == NULL) ErrorExit("Cannot create game->request.");

  game->stage_number = 0;

  game->map_width   = map_width;
  game->map_height  = map_height;
  game->width       = width;
  game->height      = height;
  game->peeping     = peeping;
  game->backward    = backward;
  game->interval    = interval_time;
  game->timeout     = timeout_time;
  game->max_players = max_players;

  return (game);
}

/*---------------------------------------------------------------------------*/
/* ֥Ȥκ                                                        */
/*---------------------------------------------------------------------------*/

Game GameDestroy(Game game)
{
  if (game == NULL) return (NULL);

  if (game->stage) StageDestroy(game->stage);

  /* tank_list  game->stage ȶͭƤΤǡ*/
  /* game->stage ˳뤳 */
  if (game->player_list) ObjList_Destroy(game->player_list);

  if (game->request) PathDestroy(game->request);

  free(game);

  return (NULL);
}

/*---------------------------------------------------------------------------*/
/* Υ                                                          */
/*---------------------------------------------------------------------------*/

#define BUFFER_SIZE (PLAYER_NAME_MAX_LENGTH + 32)

int GameStart(Game game, int silent)
{
  Path path;
  char player_name[BUFFER_SIZE];
  unsigned short int command;
  unsigned short int length;
  Player player;
  int i;
  ObjListData current;
  int ret;

  while (!GameIsOver(game)) {

    if (ObjList_GetLength(game->player_list) < 1)
      game->stage_number = 0;

    /* 饤Ȥ³׵μդ */
    while ((ObjList_GetLength(game->player_list) < PLAYER_NUMBER_MIN)
	   || PathCheckRecvState(game->request)) {
      i = 1;
      while(!PathCheckRecvState(game->request)) {
	if (!silent && !(i % 5)) printf("Waiting accept from clients...\n");
	i++;
	sleep(1);
      }

      path = PathCreateToCommunicateWithClient(game->request,
					       SERVER_SEND_BUFFER_SIZE,
					       SERVER_RECV_BUFFER_SIZE);

      while (!PathCheckRecvState(path)) { usleep(10000); }
      ret = PathCheckData(path, &command, &length);

      if (ret && (command == JOIN_REQUEST)) {

	PathRecvStringData(path, &command, player_name, BUFFER_SIZE);

	if (ObjList_GetLength(game->player_list) >= game->max_players) {
	  PathSendNoneData(path, MANY_PLAYERS);
	  PathFlushSendBuffer(path);
	  path = PathDestroy(path);
	} else {
	  player = PlayerCreate(player_name, path, PLAYER_WAIT);
	  ObjList_InsertObjToEnd(game->player_list, player,
				 (ObjDestructor)PlayerDestroy);
	  PathSendShortShortShortShortShortData(PlayerGetPath(player),
						JOIN_ACCEPT,
						game->map_width,
						game->map_height,
						game->width,
						game->height,
						game->peeping);
	  PathFlushSendBuffer(PlayerGetPath(player));
	}
      } else {
	path = PathDestroy(path);
      }
    }

    game->stage = StageCreate(game->stage_number, game->player_list,
			      game->map_width,
			      game->map_height,
			      game->width,
			      game->height,
			      game->peeping,
			      game->backward,
			      game->interval,
			      game->timeout,
			      game->max_players);
    if (game->stage == NULL) ErrorExit("Cannot create game->stage.");

    StageStartAsServer(game->stage, game->player_list, game->request);

    game->stage = StageDestroy(game->stage);

    /*
     * Player ξ֤ν
     */
    for (current = ObjList_GetStart(game->player_list);
	 !ObjList_IsEndEdge(game->player_list, current);
	 current = ObjListData_GetNext(current)) {
      player = (Player)ObjListData_GetObj(current);
      switch (PlayerGetState(player)) {
      case PLAYER_WAIT :
	break;
      case PLAYER_ACTIVE :
	break;
      case PLAYER_DESTROY : /* Destroy ֤ Player  */
	current = ObjList_DeleteObjToPrev(game->player_list, current);
	break;
      default :
      }
    }

    (game->stage_number)++;
  }

  return (0);
}

/*****************************************************************************/
/* ޤ                                                                  */
/*****************************************************************************/

/*****************************************************************************/
/* End of File.                                                              */
/*****************************************************************************/
