/*****************************************************************************/
/* N.A.D.A.R. - A network tank battle game.                                  */
/*                                                                           */
/* N.A.D.A.R. Copyright (c) 2000-2001 Sakai Hiroaki.                         */
/* All Rights Reserved.                                                      */
/*                                                                           */
/* N.A.D.A.R. is a network tank battle game. You get into a tank,            */
/* operate it, and destroy other tanks to survive!                           */
/*****************************************************************************/

/*****************************************************************************/
/* Map 饹(ޥåѤΥ饹)Υ饤֥                            */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                   */
/*****************************************************************************/

#include "MapP.h"

/*****************************************************************************/
/* ؿ                                                                */
/*****************************************************************************/

/*===========================================================================*/
/* MapChange 饹˴ؤ                                              */
/*===========================================================================*/

static MapChange MapChangeCreate(int x, int y, int old_state, int new_state)
{
  MapChange map_change;

  map_change = (MapChange)malloc(sizeof(_MapChange));
  if (map_change == NULL) return (NULL);

  map_change->x = x;
  map_change->y = y;
  map_change->old_state = old_state;
  map_change->new_state = new_state;

  return (map_change);
}

static MapChange MapChangeDestroy(MapChange map_change)
{
  if (map_change == NULL) return (NULL);
  free(map_change);
  return (NULL);
}

#if 0
static int MapChangeGetX(MapChange mc) { return (mc->x); }
static int MapChangeGetY(MapChange mc) { return (mc->y); }
static int MapChangeGetOldState(MapChange mc) { return (mc->old_state); }
static int MapChangeGetNewState(MapChange mc) { return (mc->new_state); }
#endif

/*===========================================================================*/
/* ޥå⤫ɤȽ                                                    */
/*===========================================================================*/

int MapIsInside(Map map, int x, int y)
{
  return (x >= 0 && x < map->width && y >= 0 && y < map->height);
}

/*===========================================================================*/
/* ޥåפξ֤μ                                                  */
/*===========================================================================*/

/*---------------------------------------------------------------------------*/
/* ޥåפξ֤μ                                                        */
/*---------------------------------------------------------------------------*/

int MapGetState(Map map, int x, int y)
{
  if (!MapIsInside(map, x, y)) return (-1);
  return (*(map->state + y * map->width + x));
}

/*---------------------------------------------------------------------------*/
/* ޥåפξ֤                                                        */
/*---------------------------------------------------------------------------*/

int MapSetState(Map map, int x, int y, int state)
{
  if (!MapIsInside(map, x, y)) return (-1);
  return (*(map->state + y * map->width + x) = state);
}

/*---------------------------------------------------------------------------*/
/* ޥåפξ֤ѹ                                                        */
/*---------------------------------------------------------------------------*/

int MapChangeState(Map map, int x, int y, int new_state)
{
  MapChange change;
  int old_state;
  ObjListData current;

  if (!MapIsInside(map, x, y)) return (-1);

  old_state = MapGetState(map, x, y);
  MapSetState(map, x, y, new_state);

  /* ƱξѹƤ򤬤ä顤 */

  for (current = ObjList_GetStart(map->changes);
       !ObjList_IsEndEdge(map->changes, current);
       current = ObjListData_GetNext(current)) {
    change = (MapChange)ObjListData_GetObj(current);
    if ((change->x == x) && (change->y == y)) {
      old_state = change->old_state;
      current = ObjList_DeleteObjToPrev(map->changes, current);
    }
  }

  /* ѹɲ */
  if (old_state != new_state) {
    ObjList_InsertObjToEnd(map->changes,
			   MapChangeCreate(x, y, old_state, new_state),
			   (ObjDestructor)MapChangeDestroy);
  }

  return (new_state);
}

/*===========================================================================*/
/* ޥåפν                                                            */
/*===========================================================================*/

int MapAllClean(Map map)
{
  int i;

  for (i = 0; i < (map->width) * (map->height); i++)
    map->state[i] = 0;

  while (!ObjList_IsEmpty(map->changes)) {
    ObjList_DeleteObjFromStart(map->changes);
  }

  return (0);
}

/*===========================================================================*/
/* ޥåפκ                                                              */
/*===========================================================================*/

static int GetCityState() /* ԻԤΥ饯Ǽ */
{ return (GetRandomValue(MAP_CITY_MAX  -MAP_CITY_MIN  +1) + MAP_CITY_MIN); }
static int GetBrickState() /* Υ饯Ǽ */
{ return (GetRandomValue(MAP_BRICK_MAX -MAP_BRICK_MIN +1) + MAP_BRICK_MIN); }
static int GetForestState() /* Υ饯Ǽ */
{ return (GetRandomValue(MAP_FOREST_MAX-MAP_FOREST_MIN+1) + MAP_FOREST_MIN); }

/*---------------------------------------------------------------------------*/
/* Գ1                                                                   */
/*---------------------------------------------------------------------------*/

static int MapMakeCity1(Map map)
{
  int i, j, n, x, y, vx, vy;

  n = GetRandomValue(80) + 20;
  n = n * map->width / 40 * map->height / 40;

  for (i = 0; i < n; i++) {
    x = GetRandomValue(map->width);
    y = GetRandomValue(map->height);
    vx = vy = 0;
    if (GetRandomValue(2)) vx = GetRandomValue(2) * 2 - 1;
    else                   vy = GetRandomValue(2) * 2 - 1;
    for (j = 0; j < 10; j++) {
      if (!MapIsInside(map, x, y)) break;
      MapChangeState(map, x, y, GetCityState());
      x += vx;
      y += vy;
    }
  }
  return (0);
}

/*---------------------------------------------------------------------------*/
/* Գ2                                                                   */
/*---------------------------------------------------------------------------*/

static int MapMakeCity2(Map map)
{
  int i, j, n, x, y, vx, vy;

  n = GetRandomValue(100) + 60;
  n = n * map->width / 40 * map->height / 40;

  for (i = 0; i < n; i++) {
    x = GetRandomValue(map->width);
    y = GetRandomValue(map->height);
    vx = vy = 0;
    if (GetRandomValue(2)) vx = GetRandomValue(2) * 2 - 1;
    else                   vy = GetRandomValue(2) * 2 - 1;
    for (j = 0; j < 5; j++) {
      if (!MapIsInside(map, x, y)) break;
      MapChangeState(map, x, y, GetCityState());
      x += vx;
      y += vy;
    }
  }
  return (0);
}

/*---------------------------------------------------------------------------*/
/* Գ3                                                                   */
/*---------------------------------------------------------------------------*/

static int MapMakeCity3(Map map)
{
  int i, j, n, x, y, vx, vy;

  n = GetRandomValue(100) + 40;
  n = n * map->width / 40 * map->height / 40;

  for (i = 0; i < n; i++) {
    x = GetRandomValue(map->width / 4) * 4 + 2;
    y = GetRandomValue(map->height/ 4) * 4 + 2;
    vx = vy = 0;
    if (GetRandomValue(2)) vx = GetRandomValue(2) * 2 - 1;
    else                   vy = GetRandomValue(2) * 2 - 1;
    for (j = 0; j < 5; j++) {
      if (!MapIsInside(map, x, y)) break;
      MapChangeState(map, x, y, GetCityState());
      x += vx;
      y += vy;
    }
  }
  return (0);
}

/*---------------------------------------------------------------------------*/
/* Գ4                                                                   */
/*---------------------------------------------------------------------------*/

static int MapMakeCity4(Map map)
{
  int i, j, n, x, y, vx, vy;

  n = GetRandomValue(40) + 10;
  n = n * map->width / 40 * map->height / 40;

  for (i = 0; i < n; i++) {
    x = GetRandomValue(map->width  - 1);
    y = GetRandomValue(map->height - 1);
    vx = vy = 0;
    if (GetRandomValue(2)) vx = GetRandomValue(2) * 2 - 1;
    else                   vy = GetRandomValue(2) * 2 - 1;
    for (j = 0; j < 5; j++) {
      if (MapIsInside(map,x  ,y  )) MapChangeState(map,x  ,y  ,GetCityState());
      if (MapIsInside(map,x+1,y  )) MapChangeState(map,x+1,y  ,GetCityState());
      if (MapIsInside(map,x  ,y+1)) MapChangeState(map,x  ,y+1,GetCityState());
      if (MapIsInside(map,x+1,y+1)) MapChangeState(map,x+1,y+1,GetCityState());
      x += vx * 2;
      y += vy * 2;
    }
  }
  return (0);
}

/*---------------------------------------------------------------------------*/
/* Գ5                                                                   */
/*---------------------------------------------------------------------------*/

static int MapMakeCity5(Map map)
{
  int i, j, n, x, y, vx, vy;

  for (x = 0; x < map->width; x++)
    for (y = 0; y < map->height; y++)
      MapChangeState(map, x, y, GetCityState());

  n = GetRandomValue(100) + 50;
  n = n * map->width / 40 * map->height / 40;

  for (i = 0; i < n; i++) {
    x = GetRandomValue(map->width  - 1);
    y = GetRandomValue(map->height - 1);
    vx = vy = 0;
    if (GetRandomValue(2)) vx = GetRandomValue(2) * 2 - 1;
    else                   vy = GetRandomValue(2) * 2 - 1;
    for (j = 0; j < 5; j++) {
      if (MapIsInside(map,x  ,y  )) MapChangeState(map,x  ,y  , 0);
      if (MapIsInside(map,x+1,y  )) MapChangeState(map,x+1,y  , 0);
      if (MapIsInside(map,x  ,y+1)) MapChangeState(map,x  ,y+1, 0);
      if (MapIsInside(map,x+1,y+1)) MapChangeState(map,x+1,y+1, 0);
      x += vx * 2;
      y += vy * 2;
    }
  }
  return (0);
}

/*---------------------------------------------------------------------------*/
/* 1                                                                       */
/*---------------------------------------------------------------------------*/

static int MapMakeWoods1(Map map)
{
  int x, y, n, nx, ny, i;

  x = map->width  / 2;
  y = map->height / 2;

  n = GetRandomValue(200) + 300;
  n = n * map->width / 40 * map->height / 40;

  for (i = 0; i < n; i++) {
    do {
      nx = x + (GetRandomValue(3)) - 1;
      ny = y + (GetRandomValue(3)) - 1;
    } while (!MapIsInside(map, nx, ny));
    x = nx;
    y = ny;
    MapChangeState(map, x, y, GetForestState());
  }
  return (0);
}

/*---------------------------------------------------------------------------*/
/* 2                                                                       */
/*---------------------------------------------------------------------------*/

static int MapMakeWoods2(Map map)
{
  int x, y, n, nx, ny, i, j, s;

  n = GetRandomValue(12) + 8;
  n = n * map->width / 40 * map->height / 40;

  for (j = 0; j < n; j++) {
    x = GetRandomValue(map->width);
    y = GetRandomValue(map->height);
    s = GetForestState();
    for (i = 0; i < 70; i++) {
      do {
	nx = x + (GetRandomValue(3)) - 1;
	ny = y + (GetRandomValue(3)) - 1;
      } while (!MapIsInside(map, nx, ny));
      x = nx;
      y = ny;
      MapChangeState(map, x, y, s);
    }
  }
  return (0);
}

/*---------------------------------------------------------------------------*/
/* 3                                                                       */
/*---------------------------------------------------------------------------*/

static int MapMakeWoods3(Map map)
{
  int x, y, n, dx, dy, i;

  for (x = 0; x < map->width; x++) {
    for (y = 0; y < map->height; y++) {
      MapChangeState(map, x, y, GetForestState());
    }
  }

  n = GetRandomValue(160) + 80;
  n = n * map->width / 40 * map->height / 40;

  for (i = 0; i < n; i++) {
    x = GetRandomValue(map->width);
    y = GetRandomValue(map->height);
    for (dx = -1; dx <= 1; dx++) {
      for (dy = -1; dy <= 1; dy++) {
	if (MapIsInside(map, x + dx, y + dy))
	  MapChangeState(map, x + dx, y + dy, 0);
      }
    }
  }

  return (0);
}

/*---------------------------------------------------------------------------*/
/* 4                                                                       */
/*---------------------------------------------------------------------------*/

static int MapMakeWoods4(Map map)
{
  int x, y, n, nx, ny, dx, dy, i;

  for (x = 0; x < map->width; x++)
    for (y = 0; y < map->height; y++)
      MapChangeState(map, x, y, GetForestState());

  for (dx = 1; dx <= 3; dx++) {
    for (dy = 1; dy <= 3; dy++) {
      x = map->width  * dx / 4;
      y = map->height * dy / 4;
      n = GetRandomValue(100) + 150;
      n = n * map->width / 40 * map->height / 40;
      for (i = 0; i < n; i++) {
	do {
	  nx = x + (GetRandomValue(3)) - 1;
	  ny = y + (GetRandomValue(3)) - 1;
	} while (!MapIsInside(map, nx, ny));
	x = nx;
	y = ny;
	MapChangeState(map, x, y, 0);
      }
    }
  }

  return (0);
}

/*---------------------------------------------------------------------------*/
/* 5                                                                       */
/*---------------------------------------------------------------------------*/

static int MapMakeWoods5(Map map)
{
  int x, y, n, nx, ny, i, j;

  for (x = 0; x < map->width; x++)
    for (y = 0; y < map->height; y++)
      MapChangeState(map, x, y, GetForestState());

  n = GetRandomValue(40) + 20;
  n = n * map->width / 40 * map->height / 40;

  for (j = 0; j < n; j++) {
    x = GetRandomValue(map->width);
    y = GetRandomValue(map->height);
    for (i = 0; i < 80; i++) {
      do {
	nx = x + (GetRandomValue(3)) - 1;
	ny = y + (GetRandomValue(3)) - 1;
      } while (!MapIsInside(map, nx, ny));
      x = nx;
      y = ny;
      MapChangeState(map, x, y, 0);
    }
  }
  return (0);
}

/*---------------------------------------------------------------------------*/
/* 6                                                                       */
/*---------------------------------------------------------------------------*/

static int MapMakeWoods6(Map map)
{
  int x, y, n, i, s;

  x = GetRandomValue(map->width);
  y = GetRandomValue(map->height);

  n = GetRandomValue(300) + 150;
  n = n * map->width / 40 * map->height / 40;

  s = GetForestState();
  for (i = 0; i < n; i++) {
    x += (GetRandomValue(3)) - 1;
    y += (GetRandomValue(3)) - 1;
    if (!MapIsInside(map, x, y) || MapGetState(map, x, y) != 0) {
      x = GetRandomValue(map->width);
      y = GetRandomValue(map->height);
      s = GetForestState();
    }
    MapChangeState(map, x, y, s);
  }
  return (0);
}

/*---------------------------------------------------------------------------*/
/* 7                                                                       */
/*---------------------------------------------------------------------------*/

static int MapMakeWoods7(Map map)
{
  int x, y, n, dx, dy, vx, vy, i, j, s;

  x = GetRandomValue(map->width);
  y = GetRandomValue(map->height);

  n = GetRandomValue(10) + 10;
  n = n * map->width / 40 * map->height / 40;

  for (i = 0; i < n; i++) {
    do {
      vx = GetRandomValue(7) - 3;
      vy = GetRandomValue(7) - 3;
    } while (vx*vx + vy*vy < 5 || vx*vx + vy*vy > 10 ||
	     !MapIsInside(map, x + vx, y + vy));
    for (j = 0; j < 10; j++) {
      if (!MapIsInside(map, x + vx, y + vy)) break;
      x += vx;
      y += vy;
      s = GetForestState();
      for (dx = -1; dx <= 1; dx++)
	for (dy = -1; dy <= 1; dy++)
	  if (MapIsInside(map, x+dx, y+dy))
	    MapChangeState(map, x+dx, y+dy, s);
    }
  }
  return (0);
}

/*---------------------------------------------------------------------------*/
/* 8                                                                       */
/*---------------------------------------------------------------------------*/

static int MapMakeWoods8(Map map)
{
  int x, y, n, dx, dy, vx, vy, i, j;

  for (x = 0; x < map->width; x++)
    for (y = 0; y < map->height; y++)
      MapChangeState(map, x, y, GetForestState());

  x = GetRandomValue(map->width);
  y = GetRandomValue(map->height);

  n = GetRandomValue(15) + 20;
  n = n * map->width / 40 * map->height / 40;

  for (i = 0; i < n; i++) {
    do {
      vx = GetRandomValue(7) - 3;
      vy = GetRandomValue(7) - 3;
    } while (vx*vx + vy*vy < 5 || vx*vx + vy*vy > 10 ||
	     !MapIsInside(map, x + vx, y + vy));
    for (j = 0; j < 10; j++) {
      if (!MapIsInside(map, x + vx, y + vy)) break;
      x += vx;
      y += vy;
      for (dx = -1; dx <= 1; dx++)
	for (dy = -1; dy <= 1; dy++)
	  if (MapIsInside(map, x+dx, y+dy))
	    MapChangeState(map, x+dx, y+dy, 0);
    }
  }
  return (0);
}

/*---------------------------------------------------------------------------*/
/* 9                                                                       */
/*---------------------------------------------------------------------------*/

static int MapMakeWoods9(Map map)
{
  int x, y, n, dx, dy, vx, vy, i;

  for (x = 0; x < map->width; x++)
    for (y = 0; y < map->height; y++)
      MapChangeState(map, x, y, GetForestState());

  x = GetRandomValue(map->width);
  y = GetRandomValue(map->height);

  n = GetRandomValue(10) + 15;
  n = n * map->width / 40 * map->height / 40;

  for (i = 0; i < n; i++) {
    do {
      vx = GetRandomValue(7) - 3;
      vy = GetRandomValue(7) - 3;
    } while (vx*vx + vy*vy < 5 || vx*vx + vy*vy > 10 ||
	     !MapIsInside(map, x + vx, y + vy));
    while (MapIsInside(map, x + vx, y + vy)) {
      x += vx;
      y += vy;
      for (dx = -1; dx <= 1; dx++)
	for (dy = -1; dy <= 1; dy++)
	  if (MapIsInside(map, x+dx, y+dy))
	    MapChangeState(map, x+dx, y+dy, 0);
    }
  }
  return (0);
}

/*---------------------------------------------------------------------------*/
/* 1                                                                       */
/*---------------------------------------------------------------------------*/

static int MapMakeForest1(Map map)
{
  int i, n;

  n = GetRandomValue(100) + 100;
  n = n * map->width / 40 * map->height / 40;

  for (i = 0; i < n; i++) {
    MapChangeState(map,
		   GetRandomValue(map->width),
		   GetRandomValue(map->height),
		   GetForestState());
  }
  return (0);
}

/*---------------------------------------------------------------------------*/
/* 2                                                                       */
/*---------------------------------------------------------------------------*/

static int MapMakeForest2(Map map)
{
  int i, n, x, y;

  n = GetRandomValue(200) + 150;
  n = n * map->width / 40 * map->height / 40;

  for (i = 0; i < n; i++) {
    x = (GetRandomValue(map->width + 1) +
	 GetRandomValue(map->width + 1) +
	 GetRandomValue(map->width + 1)) / 3 - 1;
    y = (GetRandomValue(map->height + 1) +
	 GetRandomValue(map->height + 1) +
	 GetRandomValue(map->height + 1)) / 3 - 1;
    MapChangeState(map, x, y, GetForestState());
  }
  return (0);
}

/*---------------------------------------------------------------------------*/
/* 3                                                                       */
/*---------------------------------------------------------------------------*/

static int MapMakeForest3(Map map)
{
  int i, n, x, y;

  n = GetRandomValue(200) + 400;
  n = n * map->width / 40 * map->height / 40;

  for (i = 0; i < n; i++) {
    x = (GetRandomValue(map->width + 1) +
	 GetRandomValue(map->width + 1) +
	 GetRandomValue(map->width + 1)) / 3 - 1;
    y = (GetRandomValue(map->height + 1) +
	 GetRandomValue(map->height + 1) +
	 GetRandomValue(map->height + 1)) / 3 - 1;
    x = (x + map->width  / 2) % map->width;
    y = (y + map->height / 2) % map->height;
    MapChangeState(map, x, y, GetForestState());
  }
  return (0);
}

/*---------------------------------------------------------------------------*/
/* ڥ1                                                               */
/*---------------------------------------------------------------------------*/

static int MapMakeSpecial1(Map map)
{
  int i, n, x, y, dx, dy;

  n = GetRandomValue(100) + 20;
  n = n * map->width / 40 * map->height / 40;

  for (i = 0; i < n; i++) {
    x = GetRandomValue(map->width);
    y = GetRandomValue(map->height);
    for (dx = -1; dx <= 1; dx++) {
      for (dy = -1; dy <= 1; dy++) {
	if (MapIsInside(map, x + dx, y + dy))
	  MapChangeState(map, x + dx, y + dy, GetCityState());
      }
    }
  }
  return (0);
}

/*---------------------------------------------------------------------------*/
/* ڥ2                                                               */
/*---------------------------------------------------------------------------*/

static int MapMakeSpecial2(Map map)
{
  int i, n, x, y, dx, dy, s0, s1;

  n = GetRandomValue(80) + 20;
  n = n * map->width / 40 * map->height / 40;

  for (i = 0; i < n; i++) {
    x = GetRandomValue(map->width);
    y = GetRandomValue(map->height);
    s0 = GetRandomValue(3);
    s1 = GetRandomValue(3);
    for (dx = -s0; dx <= s1; dx++) {
      for (dy = -s0; dy <= s1; dy++) {
	if (MapIsInside(map, x + dx, y + dy))
	  MapChangeState(map, x + dx, y + dy, GetCityState());
      }
    }
  }
  return (0);
}

/*---------------------------------------------------------------------------*/
/* ڥ3                                                               */
/*---------------------------------------------------------------------------*/

static int MapMakeSpecial3(Map map)
{
  int i, x, y;

  for (i = 2; i < map->width/2 && i < map->height/2; i += 3) {
    for (x = map->width/2 - i; x <= map->width/2 + i; x++) {
      MapChangeState(map, x, map->height/2 - i, GetBrickState());
      MapChangeState(map, x, map->height/2 + i, GetBrickState());
    }
    for (y = map->height/2 - i; y <= map->height/2 + i; y++) {
      MapChangeState(map, map->width/2 - i, y, GetBrickState());
      MapChangeState(map, map->width/2 + i, y, GetBrickState());
    }
  }

  return (0);
}

static int (*map_make_functions[])(Map) = {
#if 0
  MapMakeCity1,
  MapMakeCity2,
  MapMakeCity3,
  MapMakeCity4,
  MapMakeCity5,
  MapMakeWoods1,
  MapMakeWoods2,
  MapMakeWoods3,
  MapMakeWoods4,
  MapMakeWoods5,
  MapMakeWoods6,
  MapMakeWoods7,
  MapMakeWoods8,
  MapMakeWoods9,
  MapMakeForest1,
  MapMakeForest2,
  MapMakeForest3,
  MapMakeSpecial1,
  MapMakeSpecial2,
  MapMakeSpecial3
#else
  MapMakeCity1,
  MapMakeCity2,
  MapMakeWoods4,
  MapMakeSpecial1,
  MapMakeWoods2,
  MapMakeForest2,
  MapMakeWoods7,
  MapMakeCity4,
  MapMakeSpecial2,
  MapMakeWoods1,
  MapMakeWoods3,
  MapMakeCity5,
  MapMakeWoods8,
  MapMakeForest1,
  MapMakeWoods6,
  MapMakeCity3,
  MapMakeForest3,
  MapMakeWoods5,
  MapMakeWoods9,
  MapMakeSpecial3
#endif
};

static int map_make_functions_number = (int)(sizeof(map_make_functions) /
					     sizeof(map_make_functions[0]));

static int MapMake(Map map, int stage_number)
{
  int s;

  if (stage_number >= map_make_functions_number)
    s = GetRandomValue(map_make_functions_number);
  else
    s = stage_number % map_make_functions_number;

  return ((*(map_make_functions[s]))(map));
}

/*===========================================================================*/
/* ֥ȤȺ                                                  */
/*===========================================================================*/

/*---------------------------------------------------------------------------*/
/* ֥Ȥ                                                        */
/*---------------------------------------------------------------------------*/

Map MapCreate(Obj parent, int width, int height, int stage_number)
{
  Map map;

  map = (Map)malloc(sizeof(_Map));
  if (map == NULL) ErrorExit("Cannot allocate memory for map.");

  map->changes = ObjList_Create();
  if (map->changes == NULL) ErrorExit("Cannot create map->changes.");

  map->state = (int *)malloc(sizeof(int) * width * height);
  if (map->state == NULL) ErrorExit("Cannot allocate memory for map->state.");

  map->width  = width;
  map->height = height;

  MapAllClean(map);
  MapMake(map, stage_number);

  return (map);
}

/*---------------------------------------------------------------------------*/
/* ֥Ȥκ                                                        */
/*---------------------------------------------------------------------------*/

Map MapDestroy(Map map)
{
  if (!map) return (NULL);
  if (map->changes) ObjList_Destroy(map->changes);
  if (map->state) free(map->state);
  free(map);
  return (NULL);
}

/*===========================================================================*/
/* ǡ                                                            */
/*===========================================================================*/

/*---------------------------------------------------------------------------*/
/* ޥåפѹץ쥤䡼(Υץ쥤˻)                */
/*---------------------------------------------------------------------------*/

int MapSendMapDraw(Path path, MapChange map_change)
{
  int state;
  if (map_change->old_state && map_change->new_state) {
    PathSendCharCharShortShortData(path, MAP_DRAW2,
				   map_change->old_state,
				   map_change->new_state,
				   map_change->x, map_change->y);
  } else {
    /*
     * old_State == 0 ⤷ new_state == 0 ξˤϡ0 Ǥʤۤ
     * 롥ξȤ0ʤʤ
     */
    state = map_change->old_state | map_change->new_state;
    if (state) {
      PathSendCharShortShortData(path, MAP_DRAW,
				 state, map_change->x, map_change->y);
    }
  }
  return (0);
}

int MapSendChanges(Map map, Player player)
{
  MapChange map_change;

  while (!ObjList_IsEmpty(map->changes)) {
    map_change = (MapChange)ObjListData_GetObj(ObjList_GetStart(map->changes));
    MapSendMapDraw(PlayerGetPath(player), map_change);
    ObjList_DeleteObjFromStart(map->changes);
  }

  return (0);
}

/*---------------------------------------------------------------------------*/
/* ޥåפѹץ쥤䡼(ơγϻ)                  */
/*---------------------------------------------------------------------------*/

int MapSendChangesToAllPlayers(Map map, ObjList player_list)
{
  MapChange map_change;
  Player player;
  ObjListData current;

  while (!ObjList_IsEmpty(map->changes)) {
    map_change = (MapChange)ObjListData_GetObj(ObjList_GetStart(map->changes));
    for (current = ObjList_GetStart(player_list);
	 !ObjList_IsEndEdge(player_list, current);
	 current = ObjListData_GetNext(current)) {
      player = (Player)ObjListData_GetObj(current);
      if (PlayerGetState(player) != PLAYER_ACTIVE) continue;
      MapSendMapDraw(PlayerGetPath(player), map_change);
    }
    ObjList_DeleteObjFromStart(map->changes);
  }
  return (0);
}

/*---------------------------------------------------------------------------*/
/* ޥåפץ쥤䡼(ץ쥤䡼滲)                        */
/*---------------------------------------------------------------------------*/

int MapSendAllMap(Map map, Player player)
{
  int x, y, state;

  for (x = 0; x < map->width; x++) {
    for (y = 0; y < map->height; y++) {
      state = MapGetState(map, x, y);
      if (state) {
	PathSendCharShortShortData(PlayerGetPath(player), MAP_DRAW,
				   state, x, y);
      }
    }
  }
  return (0);
}

/*****************************************************************************/
/* ޤ                                                                  */
/*****************************************************************************/

/*****************************************************************************/
/* End of File.                                                              */
/*****************************************************************************/
