/*****************************************************************************/
/* Path.c - A library for network communication.                             */
/*                                                                           */
/* Path.c Copyright (c) 2000-2001 Sakai Hiroaki.                             */
/* All Rights Reserved.                                                      */
/*****************************************************************************/

/*****************************************************************************/
/* Path 饹(̿ϩΥ饹)                                               */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                   */
/*****************************************************************************/

#include "PathP.h"
#include "etc.h"

/*
 * ǡϥե졼˳Ǽơե졼ñ̤롥ե졼ˤϡ
 * short ե졼 long ե졼Σ̤꤬롥
 * ե졼Ƭӥåܤshort/long̤ӥåܤǥǡ
 * ̵/̤ͭ롥
 * 
 * [short ե졼ࡤǡ̵](إå:Х)
 * 0 1 2 3 4 5 6 7 8
 * +-+-+-----------+
 * |0|0|type(6bit) |
 * +-+-+-----------+
 * 
 * [short ե졼ࡤǡͭ](إå:Х)
 * 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6
 * +-+-+-----------+---------------+----+----+----+
 * |0|1|type(6bit) |length(8bit)   |data|data|... |
 * +-+-+-----------+---------------+----+----+----+
 * 
 * [long ե졼ࡤǡ̵](إå:Х)
 * 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6
 * +-+-+---------------------------+
 * |1|0|type(14bit)                |
 * +-+-+---------------------------+
 * 
 * [long ե졼ࡤǡͭ](إå:Х)
 * 0 1 2 3 4 5 ...
 * +-+-+-----------+-------------+----+----+----+
 * |1|1|type(14bit)|length(16bit)|data|data|... |
 * +-+-+-----------+-------------+----+----+----+
 * 
 * length ϡǡʬĹХñ̤ɽ
 * 
 * PATH_BUFFER_ALIGNMENT ͭˤƤݤˤϡlong/short ե졼Ȥ⡤
 * ե졼νüϡ/ХåեǾˣХȶˤʤ褦
 * ѥǥ󥰤롥äХåեǤϡ
 * ե졼ƬϡˣХȥ饤ȤƤ뤳Ȥݾڤ롥
 * (ΤᡤХåեΥǡ
 *  value = *((long int *)(p->buffer))
 *  Τ褦ˤơ㥹ȤľɤळȤǽǤ)
 *
 * PATH_BUFFER_ALIGNMENT ϡǥեȤǤ̵Ȥ롥(̵̤ʥǡ̤
 * Τ) ΤȤХåեΥǡϣХȥ饤ȤƤȤ
 * ¤ʤΤǡ
 * unsigned long int tmp;
 * *((unsigned char *)(&tmp)  ) = *((unsigned char *)(p->buffer)  );
 * *((unsigned char *)(&tmp)+1) = *((unsigned char *)(p->buffer)+1);
 * *((unsigned char *)(&tmp)+2) = *((unsigned char *)(p->buffer)+2);
 * *((unsigned char *)(&tmp)+3) = *((unsigned char *)(p->buffer)+3);
 * Τ褦ˤơŪ˥ԡ褦ʺȤɬס
 * (RISCޥǤϡint * Υݥ󥿤ϣХȥ饤ȤƤ
 *  ɬפ)
 * 
 * long ե졼ΥإåϣХȤʤΤǡPATH_BUFFER_ALIGNMENT ͭ
 * Ƥݤˤϡlong ե졼ΥǡƬϣХȥ饤
 * Ƥ뤳Ȥݾڤ롥 long int ͤȤꤹΤ
 * ѤǤ롥
 * short ե졼ΥإåϣХȤʤΤǡPATH_BUFFER_ALIGNMENT ͭ
 * Ƥݤˤϡshort ե졼ΥǡƬϣХȥ饤
 * Ƥ뤳Ȥݾڤ롥 short int ͤȤꤹΤ
 * ѤǤ롥
 *
 * ե졼إåκϥ饤֥꤬ưŪ˹Ԥ
 * ХåեХåեȤˡPATH_BUFFER_ALIGNMENT ͭ
 * ƤݤˤϡϣХȥ饤ȤƤ뤳Ȥ
 * ݾڤƤ롥
 */

#ifdef PATH_BUFFER_ALIGNMENT
#define ALIGNMENT2(x) ((((x)+1)/2)*2) /* ХȶΥ饤 */
#define ALIGNMENT4(x) ((((x)+3)/4)*4) /* ХȶΥ饤 */
#else
#define ALIGNMENT2(x) (x)
#define ALIGNMENT4(x) (x)
#endif

/* إå֤ޥ */
#ifdef PATH_BUFFER_ALIGNMENT
#define HEADER_SIZE(type,len) (((type)==PATH_LONG_FRAME)?4:2)
#else
#define HEADER_SIZE(type,len) ((((type)==PATH_LONG_FRAME)?2:1)*((len)?2:1))
#endif

/*****************************************************************************/
/* ᥽åɤ                                                            */
/*****************************************************************************/

/*===========================================================================*/
/* ե饰λѤΥ᥽åɤ                                            */
/*===========================================================================*/

/*---------------------------------------------------------------------------*/
/* ̿ϩ³                                                          */
/*---------------------------------------------------------------------------*/

int PathIsConnected(Path p)
{ return (p->connection == CONNECTION_CONNECTED); }
int PathIsDisconnected(Path p)
{ return (p->connection == CONNECTION_DISCONNECTED); }

/*===========================================================================*/
/* ֥Ȥ                                                        */
/*===========================================================================*/

static Path PathNew(int fd, int send_buffer_size, int recv_buffer_size,
		    int type)
{
  Path path;

  path = (Path)malloc(sizeof(_Path));

  path->sock = NULL;
  path->file_descriptor = fd;
  path->connection      = CONNECTION_CONNECTED;
  FD_ZERO(&(path->default_mask));
  FD_SET(fd, &(path->default_mask));

  /*
   * ХåեΥϣХȤ˥饤ȤƤʤȤʤ
   */
  path->send_buffer_size = ALIGNMENT4(send_buffer_size);
  path->recv_buffer_size = ALIGNMENT4(recv_buffer_size);

  path->send_buffer = (unsigned char *)malloc(path->send_buffer_size);
  path->recv_buffer = (unsigned char *)malloc(path->recv_buffer_size);

  path->send_buffer_usable = path->send_buffer;
  path->recv_buffer_start  = path->recv_buffer;
  path->recv_buffer_usable = path->recv_buffer;

  path->type = type;

  return (path);
}

Path PathCreateToAcceptRequestsFromClients(char * hostname,
					   unsigned short int port,
					   int backlog)
{
  Socket s;
  Path path;

  s = SocketCreateToAccept(hostname, port, backlog);
  if (!s) ErrorExit("Cannot create socket to accept.");

  /* accept() Ѥ Path ʤΤǡХåե */
  path = PathNew(SocketGetSocket(s), 1, 1, PATH_FOR_ACCEPT);
  if (!path) ErrorExit("Cannot create a new path.");

  path->sock = s;

  return (path);
}

Path PathCreateToCommunicateWithClient(Path path,
				       int send_buffer_size,
				       int recv_buffer_size)
{
  Path p;
  Socket s;

  s = SocketCreateByAccept(path->sock);
  if (!s) return (NULL);
  p = PathNew(SocketGetSocket(s), send_buffer_size, recv_buffer_size,
	      PATH_FOR_COMMUNICATE);
  p->sock = s;

#ifdef SOCKET_TCP_NODELAY_ON
  SocketSetTcpNodelay(p->sock, 1);
#endif

  return (p);
}

Path PathCreateToCommunicateWithServer(char * hostname,
				       unsigned short int port,
				       int send_buffer_size,
				       int recv_buffer_size)
{
  Socket s;
  Path path;

  s = SocketCreateToConnect(hostname, port);
  if (!s) return (NULL);
  path = PathNew(SocketGetSocket(s), send_buffer_size, recv_buffer_size,
		 PATH_FOR_COMMUNICATE);
  path->sock = s;

#ifdef SOCKET_TCP_NODELAY_ON
  SocketSetTcpNodelay(path->sock, 1);
#endif

  return (path);
}

/*===========================================================================*/
/* ֥Ȥξõ                                                        */
/*===========================================================================*/

Path PathDestroy(Path path)
{
  if (path == NULL) return (NULL);

  SocketDestroy(path->sock);

  if (path->send_buffer) free(path->send_buffer);
  if (path->recv_buffer) free(path->recv_buffer);

  free(path);

  return (NULL);
}

/*****************************************************************************/
/* ǡѤΥ᥽å                                                */
/*****************************************************************************/

/*===========================================================================*/
/* ǡΥ᥽åɤ                                              */
/*===========================================================================*/

/*---------------------------------------------------------------------------*/
/* char ΥǡХåե˳Ǽ                                             */
/*---------------------------------------------------------------------------*/

static int PathPutCharData(Path p, unsigned char c)
{
  *((unsigned char *)p->send_buffer_usable) = c;
  p->send_buffer_usable += 1;
  return (0);
}

/*---------------------------------------------------------------------------*/
/* short int ΥǡХåե˳Ǽ                                        */
/*---------------------------------------------------------------------------*/

static int PathPutShortData(Path p, unsigned short int s)
{
  s = htons(s);
#ifdef PATH_BUFFER_ALIGNMENT
  *((unsigned short int *)p->send_buffer_usable) = s;
#else
  *(p->send_buffer_usable    ) = *((unsigned char *)(&s)    );
  *(p->send_buffer_usable + 1) = *((unsigned char *)(&s) + 1);
#endif
  /*
   * ((unsigned short int *)p->send_buffer_usable)++;
   * Τ褦ʥɤ񤤤ƤϤʤ(ܸ C-FAQ μ 4.5 )
   */
  p->send_buffer_usable += (sizeof(unsigned short int)/sizeof(unsigned char));
  return (0);
}

/*---------------------------------------------------------------------------*/
/* long int ΥǡХåե˳Ǽ                                         */
/*---------------------------------------------------------------------------*/

static int PathPutLongData(Path p, unsigned long int s)
{
  s = htonl(s);
#ifdef PATH_BUFFER_ALIGNMENT
  *((unsigned long int *)p->send_buffer_usable) = s;
#else
  *(p->send_buffer_usable    ) = *((unsigned char *)(&s)    );
  *(p->send_buffer_usable + 1) = *((unsigned char *)(&s) + 1);
  *(p->send_buffer_usable + 2) = *((unsigned char *)(&s) + 2);
  *(p->send_buffer_usable + 3) = *((unsigned char *)(&s) + 3);
#endif
  /*
   * ((unsigned long int *)p->send_buffer_usable)++;
   * Τ褦ʥɤ񤤤ƤϤʤ(ܸ C-FAQ μ 4.5 )
   */
  p->send_buffer_usable += (sizeof(unsigned long int)/sizeof(unsigned char));
  return (0);
}

/*---------------------------------------------------------------------------*/
/* Хʥ                                                            */
/*---------------------------------------------------------------------------*/

static int PathSend(Path p, char * s, int len)
{
#if 0
  int i;
#endif

  if (len <= 0) return (0);
  len = send(p->file_descriptor, s, len, 0);

#if 0
  for (i = 0; i < len; i++) printf("%c", s[i]);
  fflush(stdout);
#endif

  if (len < 0) return (-1);
  return (len);
}

/*---------------------------------------------------------------------------*/
/* ХåեΥեå                                                  */
/*---------------------------------------------------------------------------*/

int PathFlushSendBuffer(Path p)
{
  int len;
  len = PathSend(p, p->send_buffer, p->send_buffer_usable - p->send_buffer);
  p->send_buffer_usable = p->send_buffer;
  return (len);
}

/*---------------------------------------------------------------------------*/
/* ե졼Υإå                                                    */
/*---------------------------------------------------------------------------*/

static char * PathMakeHeader(Path p,
			     int long_flag,
			     unsigned short int type,
			     unsigned short int length)
{
  int len;
  unsigned char c1;
  unsigned char c2;
  unsigned char * old;

  /* ե졼Υ׻ */
  len = HEADER_SIZE(long_flag, length) + length;
  len = ALIGNMENT4(len); /* ХȤ˥饤 */

  if (p->send_buffer_usable + len > p->send_buffer + p->send_buffer_size) {
    /* ХåեСեƤޤ硥Хåեեå夹 */
    if (PathFlushSendBuffer(p) < 0) ErrorExit("send error.");
    if (p->send_buffer_usable + len > p->send_buffer + p->send_buffer_size) {
      /*
       * Ǥ⥪СեƤޤʤ顤ХåեΤǡ
       * Error λ(ưŪ˥Хåե realloc() Ƥ⤤ɡ
       * ХФȤ˸ĤˤʤꤽʤΤǤȤ)
       */
      ErrorExit("send buffer is too small.");
    } else {
#ifndef PATH_SEND_BUFFER_AUTO_FLUSH
    /*
     * ߥ󥰤褯եåԤäƤʤȤȤʤΤǡError λ
     * ΥåФȤϡХåե˥ǡ򤿤᤿ˡ
     * Ŭڤʥߥ󥰤ǥեåԤȡ
     */
    ErrorExit("you need to flush buffer.");
#endif
    }
  }

  old = p->send_buffer_usable;

  switch (long_flag) {
  case PATH_LONG_FRAME :
    if (type & 0xc000) ErrorExit("Invalid data type.");
    /* if (length > 0xffff) ErrorExit("Data length is too long."); */

    /*
     * Хȥ¸ˤʤʤ褦ˡƬХȤȼХȤ
     * ʬ롥
     */
    c1 = (type >> 8) & 0xff;
    c2 = type & 0xff;

    /*
     * long եޥåȤξˤϡǾ̥ӥåȤΩƤ롥
     * length ξ(ǡ)ϡӥåܤΩƤ롥
     */
    if (length) c1 = c1 | 0xc0;
    else c1 = c1 | 0x80;
    PathPutCharData(p, c1);
    PathPutCharData(p, c2);

    if (length)
      PathPutShortData(p, length);

    break;

  case PATH_SHORT_FRAME :
    if (type >= 0x40) ErrorExit("Invalid data type.");
    if (length > 0xff) ErrorExit("Data length is too long.");

    /* short եޥåȤξˤϡǾ̥ӥåȤΩƤʤ */
    c1 = type;
    if (length) c1 = c1 | 0x40;

    PathPutCharData(p, c1);
    if (length) PathPutCharData(p, length & 0xff);

    break;

  default :
    ErrorExit("Unknown data format.");
  }

  /*
   * ե졼Ƭɥ쥹(饤ȤƤ)֤
   */
  return(old + len);
}

/*---------------------------------------------------------------------------*/
/* ʤǡ(type )                                   */
/*---------------------------------------------------------------------------*/

int PathSendNoneData(Path p, unsigned short int type)
{
  char * next_start_point;
  unsigned short int len;

  len = 0;

  next_start_point = PathMakeHeader(p, PATH_SHORT_FRAME, type, len);

  /* ХȤ˥饤Ȥ */
  p->send_buffer_usable = next_start_point;

  return (0);
}

/*---------------------------------------------------------------------------*/
/* short Υǡ                                                      */
/*---------------------------------------------------------------------------*/

int PathSendShortData(Path p, unsigned short int type,
		      unsigned short int s1)
{
  char * next_start_point;
  unsigned short int len;

  len = sizeof(s1);

  next_start_point = PathMakeHeader(p, PATH_SHORT_FRAME, type, len);

  PathPutShortData(p, s1);

  /* ХȤ˥饤Ȥ */
  p->send_buffer_usable = next_start_point;

  return (0);
}

/*---------------------------------------------------------------------------*/
/* short, short Υǡ                                               */
/*---------------------------------------------------------------------------*/

int PathSendShortShortData(Path p, unsigned short int type,
			   unsigned short int s1,
			   unsigned short int s2)
{
  char * next_start_point;
  unsigned short int len;

  len = sizeof(s1) + sizeof(s2);

  next_start_point = PathMakeHeader(p, PATH_SHORT_FRAME, type, len);

  PathPutShortData(p, s1);
  PathPutShortData(p, s2);

  /* ХȤ˥饤Ȥ */
  p->send_buffer_usable = next_start_point;

  return (0);
}

/*---------------------------------------------------------------------------*/
/* short, short, short Υǡ                                        */
/*---------------------------------------------------------------------------*/

int PathSendShortShortShortData(Path p, unsigned short int type,
				unsigned short int s1,
				unsigned short int s2,
				unsigned short int s3)
{
  char * next_start_point;
  unsigned short int len;

  len = sizeof(s1) + sizeof(s2) + sizeof(s3);

  next_start_point = PathMakeHeader(p, PATH_SHORT_FRAME, type, len);

  PathPutShortData(p, s1);
  PathPutShortData(p, s2);
  PathPutShortData(p, s3);

  /* ХȤ˥饤Ȥ */
  p->send_buffer_usable = next_start_point;

  return (0);
}

/*---------------------------------------------------------------------------*/
/* short, short, short, short Υǡ                                 */
/*---------------------------------------------------------------------------*/

int PathSendShortShortShortShortData(Path p, unsigned short int type,
				     unsigned short int s1,
				     unsigned short int s2,
				     unsigned short int s3,
				     unsigned short int s4)
{
  char * next_start_point;
  unsigned short int len;

  len = sizeof(s1) + sizeof(s2) + sizeof(s3) + sizeof(s4);

  next_start_point = PathMakeHeader(p, PATH_SHORT_FRAME, type, len);

  PathPutShortData(p, s1);
  PathPutShortData(p, s2);
  PathPutShortData(p, s3);
  PathPutShortData(p, s4);

  /* ХȤ˥饤Ȥ */
  p->send_buffer_usable = next_start_point;

  return (0);
}

/*---------------------------------------------------------------------------*/
/* short, short, short, short, short Υǡ                          */
/*---------------------------------------------------------------------------*/

int PathSendShortShortShortShortShortData(Path p, unsigned short int type,
					  unsigned short int s1,
					  unsigned short int s2,
					  unsigned short int s3,
					  unsigned short int s4,
					  unsigned short int s5)
{
  char * next_start_point;
  unsigned short int len;

  len = sizeof(s1) + sizeof(s2) + sizeof(s3) + sizeof(s4) + sizeof(s5);

  next_start_point = PathMakeHeader(p, PATH_SHORT_FRAME, type, len);

  PathPutShortData(p, s1);
  PathPutShortData(p, s2);
  PathPutShortData(p, s3);
  PathPutShortData(p, s4);
  PathPutShortData(p, s5);

  /* ХȤ˥饤Ȥ */
  p->send_buffer_usable = next_start_point;

  return (0);
}

/*---------------------------------------------------------------------------*/
/* long Υǡ                                                       */
/*---------------------------------------------------------------------------*/

int PathSendLongData(Path p, unsigned short int type,
		     unsigned long int s1)
{
  char * next_start_point;
  unsigned short int len;

  len = sizeof(s1);

  /*
   * PATH_BUFFER_ALIGNMENT ƤˤϡХȥ饤
   * ɬפΤǡlong frame ǤʤФʤʤ
   */
#ifdef PATH_BUFFER_ALIGNMENT
  next_start_point = PathMakeHeader(p, PATH_LONG_FRAME, type, len);
#else
  next_start_point = PathMakeHeader(p, PATH_SHORT_FRAME, type, len);
#endif

  PathPutLongData(p, s1);

  /* ХȤ˥饤Ȥ */
  p->send_buffer_usable = next_start_point;

  return (0);
}

/*---------------------------------------------------------------------------*/
/* long, long Υǡ                                                 */
/*---------------------------------------------------------------------------*/

int PathSendLongLongData(Path p, unsigned short int type,
			 unsigned long int s1,
			 unsigned long int s2)
{
  char * next_start_point;
  unsigned short int len;

  len = sizeof(s1) + sizeof(s2);

  /*
   * PATH_BUFFER_ALIGNMENT ƤˤϡХȥ饤
   * ɬפΤǡlong frame ǤʤФʤʤ
   */
#ifdef PATH_BUFFER_ALIGNMENT
  next_start_point = PathMakeHeader(p, PATH_LONG_FRAME, type, len);
#else
  next_start_point = PathMakeHeader(p, PATH_SHORT_FRAME, type, len);
#endif

  PathPutLongData(p, s1);
  PathPutLongData(p, s2);

  /* ХȤ˥饤Ȥ */
  p->send_buffer_usable = next_start_point;

  return (0);
}

/*---------------------------------------------------------------------------*/
/* char, char, char, char Υǡ                                     */
/*---------------------------------------------------------------------------*/

int PathSendCharCharCharCharData(Path p, unsigned short int type,
				 unsigned char c1, unsigned char c2,
				 unsigned char c3, unsigned char c4)
{
  char * next_start_point;
  unsigned short int len;

  len = sizeof(c1) + sizeof(c2) + sizeof(c3) + sizeof(c4);

  next_start_point = PathMakeHeader(p, PATH_SHORT_FRAME, type, len);

  PathPutCharData(p, c1);
  PathPutCharData(p, c2);
  PathPutCharData(p, c3);
  PathPutCharData(p, c4);

  /* ХȤ˥饤Ȥ */
  p->send_buffer_usable = next_start_point;

  return (0);
}

/*---------------------------------------------------------------------------*/
/* char, char, short, short Υǡ                                   */
/*---------------------------------------------------------------------------*/

int PathSendCharCharShortShortData(Path p, unsigned short int type,
				   unsigned char c1,
				   unsigned char c2,
				   unsigned short int s1,
				   unsigned short int s2)
{
  char * next_start_point;
  unsigned short int len;

  len = sizeof(c1) + sizeof(c2) + sizeof(s1) + sizeof(s2);

  next_start_point = PathMakeHeader(p, PATH_SHORT_FRAME, type, len);

  PathPutCharData(p, c1);
  PathPutCharData(p, c2);
  PathPutShortData(p, s1);
  PathPutShortData(p, s2);

  /* ХȤ˥饤Ȥ */
  p->send_buffer_usable = next_start_point;

  return (0);
}

/*---------------------------------------------------------------------------*/
/* char, short, short Υǡ                                         */
/*---------------------------------------------------------------------------*/

int PathSendCharShortShortData(Path p, unsigned short int type,
			       unsigned char c1,
			       unsigned short int s1,
			       unsigned short int s2)
{
  /*
   * PATH_BUFFER_ALIGNMENT Ƥˤϡ饤
   * ɬפΤǡ
   * PathSendCharCharShortShortData()롥
   */
#ifdef PATH_BUFFER_ALIGNMENT
  return (PathSendCharCharShortShortData(p, type, c1, 0, s1, s2));
#else
  char * next_start_point;
  unsigned short int len;

  len = sizeof(c1) + sizeof(s1) + sizeof(s2);

  next_start_point = PathMakeHeader(p, PATH_SHORT_FRAME, type, len);

  PathPutCharData(p, c1);
  PathPutShortData(p, s1);
  PathPutShortData(p, s2);

  p->send_buffer_usable = next_start_point;

  return (0);
#endif
}

/*---------------------------------------------------------------------------*/
/* ʸǡ                                                        */
/*---------------------------------------------------------------------------*/

int PathSendStringData(Path p, unsigned short int type, char * string)
{
  char * next_start_point;
  unsigned short int len;
  unsigned short int str_length;

  str_length = strlen(string);
  /* len = sizeof(str_length) + str_length; */
  len = str_length; /* ʸĹɬפ̵ */

  next_start_point = PathMakeHeader(p, PATH_LONG_FRAME, type, len);

  /* PathPutShortData(p, str_length); */
  /* '\0' ɬפ̵ */
  memcpy(p->send_buffer_usable, string, str_length);

  /* ХȤ˥饤Ȥ */
  p->send_buffer_usable = next_start_point;

  return (0);
}

/*---------------------------------------------------------------------------*/
/* short, string Υǡ                                              */
/*---------------------------------------------------------------------------*/

int PathSendShortStringData(Path p, unsigned short int type,
			    unsigned short int s1,
			    char * string)
{
  char * next_start_point;
  unsigned short int len;
  unsigned short int str_length;

  str_length = strlen(string);
  len = sizeof(s1) + str_length;

  next_start_point = PathMakeHeader(p, PATH_LONG_FRAME, type, len);

  PathPutShortData(p, s1);

  /* PathPutShortData(p, str_length); */
  /* '\0' ɬפ̵ */
  memcpy(p->send_buffer_usable, string, str_length);

  /* ХȤ˥饤Ȥ */
  p->send_buffer_usable = next_start_point;

  return (0);
}

/*---------------------------------------------------------------------------*/
/* char, char, short, short, string Υǡ                           */
/*---------------------------------------------------------------------------*/

int PathSendCharCharShortShortStringData(Path p, unsigned short int type,
					 unsigned char c1,
					 unsigned char c2,
					 unsigned short int s1,
					 unsigned short int s2,
					 char * string)
{
  char * next_start_point;
  unsigned short int len;
  unsigned short int str_length;

  str_length = strlen(string);
  len = sizeof(c1) + sizeof(c2) + sizeof(s1) + sizeof(s2) + str_length;

  next_start_point = PathMakeHeader(p, PATH_LONG_FRAME, type, len);

  PathPutCharData(p, c1);
  PathPutCharData(p, c2);
  PathPutShortData(p, s1);
  PathPutShortData(p, s2);

  /* PathPutShortData(p, str_length); */
  /* '\0' ɬפ̵ */
  memcpy(p->send_buffer_usable, string, str_length);

  /* ХȤ˥饤Ȥ */
  p->send_buffer_usable = next_start_point;

  return (0);
}

/*---------------------------------------------------------------------------*/
/* short, short, short, short, string Υǡ                         */
/*---------------------------------------------------------------------------*/

int PathSendShortShortShortShortStringData(Path p, unsigned short int type,
					   unsigned short int s1,
					   unsigned short int s2,
					   unsigned short int s3,
					   unsigned short int s4,
					   char * string)
{
  char * next_start_point;
  unsigned short int len;
  unsigned short int str_length;

  str_length = strlen(string);
  len = sizeof(s1) + sizeof(s2) + sizeof(s3) + sizeof(s4) + str_length;

  next_start_point = PathMakeHeader(p, PATH_LONG_FRAME, type, len);

  PathPutShortData(p, s1);
  PathPutShortData(p, s2);
  PathPutShortData(p, s3);
  PathPutShortData(p, s4);

  /* PathPutShortData(p, str_length); */
  /* '\0' ɬפ̵ */
  memcpy(p->send_buffer_usable, string, str_length);

  /* ХȤ˥饤Ȥ */
  p->send_buffer_usable = next_start_point;

  return (0);
}

/*===========================================================================*/
/* ǡμΥ᥽åɤ                                              */
/*===========================================================================*/

/*---------------------------------------------------------------------------*/
/* ѥξ֤Ĵ٤ʤ                                                */
/*---------------------------------------------------------------------------*/

/*
 * ѥ select() ơǡƤʤɤĴ٤롥
 * ǡƤ 1 ֤ǡ̵ä 0 ֤
 */

int PathCheckRecvState(Path p)
{
  fd_set recv_mask;
  static struct timeval timeout = {0, 0};

  recv_mask = p->default_mask;
  select(p->file_descriptor + 1, &recv_mask, NULL, NULL, &timeout);

  if (FD_ISSET(p->file_descriptor, &recv_mask)) return (1);

  return (0);
}

/*---------------------------------------------------------------------------*/
/* ХåեΥǡƬ˥եȤơ̤ΰ򹭤롥            */
/*---------------------------------------------------------------------------*/

static int PathShiftRecvBuffer(Path p)
{
#ifdef PATH_BUFFER_ALIGNMENT
  unsigned long int * p1;
  unsigned long int * p2;
#else
  unsigned char * p1;
  unsigned char * p2;
#endif

  if (p->recv_buffer == p->recv_buffer_start) return (0);

#ifdef PATH_BUFFER_ALIGNMENT
  /*
   * ХåեΥȥե졼ƬϡХȥ饤ȤƤΤǡ
   * unsigned long int ñ̤ǥեȤƤ褤
   */
  p1 = (unsigned long int *)(p->recv_buffer);
  p2 = (unsigned long int *)(p->recv_buffer_start);
  while ((unsigned char *)p2 < p->recv_buffer_usable) {
    *p1 = *p2;
    p1++;
    p2++;
  }
#else
  /*
   * ХåեΥȥե졼ƬϡХȥ饤ȤƤʤ
   */
  p1 = (unsigned char *)(p->recv_buffer);
  p2 = (unsigned char *)(p->recv_buffer_start);
  while ((unsigned char *)p2 < p->recv_buffer_usable) {
    *p1 = *p2;
    p1++;
    p2++;
  }
#endif

  p->recv_buffer_usable -= (p->recv_buffer_start - p->recv_buffer);
  p->recv_buffer_start = p->recv_buffer;

  return (0);
}

/*---------------------------------------------------------------------------*/
/* char ΥǡХåե                                           */
/*---------------------------------------------------------------------------*/

static unsigned char PathGetCharData(Path p)
{
  unsigned char c;
  c = *((unsigned char *)p->recv_buffer_start);
  p->recv_buffer_start += 1;
  return (c);
}

/*---------------------------------------------------------------------------*/
/* short int ΥǡХåե                                      */
/*---------------------------------------------------------------------------*/

static unsigned short int PathGetShortData(Path p)
{
  unsigned short int s;
#ifdef PATH_BUFFER_ALIGNMENT
  s = *((unsigned short int *)p->recv_buffer_start);
#else
  *((unsigned char *)(&s)    ) = *(p->recv_buffer_start    );
  *((unsigned char *)(&s) + 1) = *(p->recv_buffer_start + 1);
#endif
  /*
   * ((unsigned short int *)p->recv_buffer_start)++;
   * Τ褦ʥɤ񤤤ƤϤʤ(ܸ C-FAQ μ 4.5 )
   */
  p->recv_buffer_start += (sizeof(unsigned short int)/sizeof(unsigned char));
  s = ntohs(s);
  return (s);
}

/*---------------------------------------------------------------------------*/
/* long int ΥǡХåե                                       */
/*---------------------------------------------------------------------------*/

static unsigned long int PathGetLongData(Path p)
{
  unsigned long int s;
#ifdef PATH_BUFFER_ALIGNMENT
  s = *((unsigned long int *)p->recv_buffer_start);
#else
  *((unsigned char *)(&s)    ) = *(p->recv_buffer_start    );
  *((unsigned char *)(&s) + 1) = *(p->recv_buffer_start + 1);
  *((unsigned char *)(&s) + 2) = *(p->recv_buffer_start + 2);
  *((unsigned char *)(&s) + 3) = *(p->recv_buffer_start + 3);
#endif
  /*
   * ((unsigned long int *)p->recv_buffer_start)++;
   * Τ褦ʥɤ񤤤ƤϤʤ(ܸ C-FAQ μ 4.5 )
   */
  p->recv_buffer_start += (sizeof(unsigned long int)/sizeof(unsigned char));
  s = ntohl(s);
  return (s);
}

/*---------------------------------------------------------------------------*/
/* char ΥǡХåե(ݥ󥿤ʤʤ)                       */
/*---------------------------------------------------------------------------*/

static unsigned char PathPeekCharData(Path p, int offset)
{
  unsigned char c;
  c = *((unsigned char *)p->recv_buffer_start + offset);
  return (c);
}

/*---------------------------------------------------------------------------*/
/* short int ΥǡХåե(ݥ󥿤ʤʤ)                  */
/*---------------------------------------------------------------------------*/

static unsigned short int PathPeekShortData(Path p, int offset)
{
  unsigned short int s;
#ifdef PATH_BUFFER_ALIGNMENT
  s = *((unsigned short int *)p->recv_buffer_start + offset);
#else
  offset *= sizeof(unsigned short int);
  *((unsigned char *)(&s)    ) = *(p->recv_buffer_start + offset    );
  *((unsigned char *)(&s) + 1) = *(p->recv_buffer_start + offset + 1);
#endif
  s = ntohs(s);
  return (s);
}

/*---------------------------------------------------------------------------*/
/* long int ΥǡХåե(ݥ󥿤ʤʤ)                   */
/*---------------------------------------------------------------------------*/

#if 0
static unsigned long int PathPeekLongData(Path p, int offset)
{
  unsigned long int s;
#ifdef PATH_BUFFER_ALIGNMENT
  s = *((unsigned long int *)p->recv_buffer_start + offset);
#else
  offset *= sizeof(unsigned long int);
  *((unsigned char *)(&s)    ) = *(p->recv_buffer_start + offset    );
  *((unsigned char *)(&s) + 1) = *(p->recv_buffer_start + offset + 1);
  *((unsigned char *)(&s) + 2) = *(p->recv_buffer_start + offset + 2);
  *((unsigned char *)(&s) + 3) = *(p->recv_buffer_start + offset + 3);
#endif
  s = ntohl(s);
  return (s);
}
#endif

/*---------------------------------------------------------------------------*/
/* Хʥμ                                                            */
/*---------------------------------------------------------------------------*/

static int PathRecv(Path p, char * buffer, int len)
{
#if 0
  int i;
#endif

  len = recv(p->file_descriptor, buffer, len, 0);

#if 0
  for (i = 0; i < len; i++) printf("%c", buffer[i]);
  fflush(stdout);
#endif

  if (len < 0) return (-1);
  return (len);
}

/*---------------------------------------------------------------------------*/
/* Хåեؤμ                                                          */
/*---------------------------------------------------------------------------*/

static int PathRecvToBuffer(Path p)
{
  int len;
  int size;

  if (!PathCheckRecvState(p)) return (-1);
  size = (p->recv_buffer + p->recv_buffer_size) - p->recv_buffer_usable;
  if (size <= 0) return (0);

  len = PathRecv(p, p->recv_buffer_usable, size);
  if (len < 0) return (-1);

  p->recv_buffer_usable += len;

  return (len);
}

/*---------------------------------------------------------------------------*/
/* ե졼Υإå                                                    */
/*---------------------------------------------------------------------------*/

/*
 * ХåեˤޤäƤե졼ΤƬˤե졼Υإåʬ
 * Ϥ롥
 * ե졼ΤХåե¸ߤɤ߹߲ǽʾϡ
 * type  length ͤꤷإåΥ֤
 * ե졼बǤڤƤˤϡưǼ롥
 * ɤե졼ब̵(Хåե˥ե졼ब̵äƤ
 * ڤƤơselect()ƤʤˤƤƤʤ)ˤϡ
 * 0 ֤
 */

static int PathCheckHeader(Path p,
			   unsigned short int * type,
			   unsigned short int * length)
{
  int long_flag;
  int data_flag;
  unsigned short int type_tmp = 0;
  unsigned short int length_tmp = 0;
  unsigned char c1;
  unsigned char c2;
  int already_shifted = 0;
  int already_recieved = 0;
  int len;

  /* إåʬХåե̵ˤϡ */
  if (p->recv_buffer_usable - p->recv_buffer_start <
      HEADER_SIZE(PATH_SHORT_FRAME, 0)) { /* ǤûإåĹ */

    /*
     * Хåեˤޤ̵ȤϡեȤ롥
     * (¡إåʬ٤ˤϥեȤɬפ)
     */
    if ((p->recv_buffer + p->recv_buffer_size) - p->recv_buffer_usable < 8) {
      PathShiftRecvBuffer(p);
      already_shifted = 1;
    }

    if (PathRecvToBuffer(p) < 0) return (0);
    already_recieved = 1;

    /* ǤإåʬХåե̵ˤϡǡ̵ */
    if (p->recv_buffer_usable - p->recv_buffer_start <
	HEADER_SIZE(PATH_SHORT_FRAME, 0)) return (0);
  }

  /*
   * ա إåǥǡڤƤ礬뤫⤷ʤ
   * Ȥθƥǥ󥰤뤳ȡ
   */

  /*
   * 顼ξˤϡξ֤᤹ɬפΤǡPathGetCharData() 
   * ͤƤϤʤ(PathGetCharData() ȡp->recv_buffer_start 
   * ѲƤޤΤ)
   * PathPeekCharData() ȡp->recv_buffer_start Ѳʤ
   */
  c1 = PathPeekCharData(p, 0);

  long_flag = (c1 & 0x80) ? PATH_LONG_FRAME : PATH_SHORT_FRAME;
  data_flag = (c1 & 0x40) ? 1 : 0;

  /*
   * إåʬХåեˤ뤳Ȥǧ롥
   * c1 = PathPeekCharData(p);
   * ˤꡤp->recv_buffer_start ͤϿʤǤʤΤǡ
   * (p->recv_buffer_usable - p->recv_buffer_start < HEADER_SIZE()) 
   * ʤäƤ뤫ɤå롥
   */
  if (p->recv_buffer_usable - p->recv_buffer_start <
      HEADER_SIZE(long_flag, data_flag)) {
    /* إåʬ­ʤȤˤϡ */

    /*
     * Хåեˤޤ̵ȤϡեȤ롥
     * (¡إåʬ٤ˤϥեȤɬפ)
     */
    if ((!already_shifted) &&
	((p->recv_buffer + p->recv_buffer_size) - p->recv_buffer_usable < 8)) {
      PathShiftRecvBuffer(p);
      already_shifted = 1;
    }

    if (!already_recieved) {
      if (PathRecvToBuffer(p) < 0) return (0);
    }

    /* Ǥإåʬ­ʤȤˤϡǡ̵ */
    if (p->recv_buffer_usable - p->recv_buffer_start <
	HEADER_SIZE(long_flag, data_flag)) return (0);
  }

  switch (long_flag) {

  case PATH_LONG_FRAME :

    c2 = PathPeekCharData(p, 1);
    type_tmp = (((unsigned short int)(c1 & 0x3f)) << 8) | c2;
    if (data_flag) length_tmp = PathPeekShortData(p, 1);
    else length_tmp = 0;
    break;

  case PATH_SHORT_FRAME :

    type_tmp = c1 & 0x3f;
    if (data_flag) length_tmp = PathPeekCharData(p, 1);
    else length_tmp = 0;
    break;

  default :
    ErrorExit("Unknown type.");
  }

  len = HEADER_SIZE(long_flag, data_flag) + length_tmp;
  len = ALIGNMENT4(len);

  /* ե졼ΤХåեˤ뤳Ȥå */
  if (p->recv_buffer_start + len > p->recv_buffer_usable) {
    /* ե졼ΤХåե̵ */

    /*
     * ե졼ΤХåե֤뤫ɤå롥
     */
    if (p->recv_buffer_start + len > p->recv_buffer + p->recv_buffer_size) {

      /*
       * ե졼ΤХåե꤭ʤϥեȤ롥
       */

      if (already_shifted) {
	/*
	 * Ǥ˥եȤƤˤ⤫餺ե졼ΤХåե
	 * ڤʤϡ顼λ
	 */
	ErrorExit("receive uffer is too small.");
      }

      PathShiftRecvBuffer(p);
      if (p->recv_buffer_start + len > p->recv_buffer + p->recv_buffer_size) {
	/*
	 * եȤˤ⤫餺ե졼ΤХåե꤭ʤ
	 * ˤϡ顼λ
	 */
	ErrorExit("buffer is too small.");
      }

      /*
       * եȤˤäƥХåեζ򹭤Τǡalready_received 
       * 餺롥
       */
      if (PathRecvToBuffer(p) < 0) return (0);

    } else {
      /* ե졼ΤХåե꤭ */
      if (already_recieved) return (0);
      if (PathRecvToBuffer(p) < 0) return (0);
    }

    /*
     * ǤХåե˥ե졼Τ̵ˤϡǡ̵
     * (PathShiftRecvBuffer(p)  PathRecvToBuffer(p) ˤꡤʳǤ
     * ݥ󥿤ΰ֤ѲƤΤǡʲΥݥ󥿤Ӥդ뤳ȡ
     */
    if (p->recv_buffer_start + len > p->recv_buffer_usable)
      return (0);
  }

  /*
   * ޤǤȤȤϡХåեˤϥե졼ΤǼ
   * ȤȤǤ롥
   */

  *type = type_tmp;
  *length = length_tmp;

  return (HEADER_SIZE(long_flag, data_flag));
}

/*
 * Хåեɤ߹ǡ뤫ɤĴ٤롥()
 * ɤ߹߲ǽʥǡ¸ߤϡtype  length ͤꤷ
 * 1 ̵֤ˤϡselect() Ǽǡ̵ɤưĴ١
 * Ǥ̵ 0 ֤
 */
int PathCheckData(Path p,
		  unsigned short int * type,
		  unsigned short int * length)
{
  int ret;
  unsigned short int type_tmp;
  unsigned short int length_tmp;

  ret = PathCheckHeader(p, &type_tmp, &length_tmp);
  ret = ret ? 1 : 0;
  if (ret) {
    *type = type_tmp;
    *length = length_tmp;
  }
  return (ret);
}

static char * PathReadHeader(Path p,
			     unsigned short int * type,
			     unsigned short int * length)
{
  unsigned short int type_tmp;
  unsigned short int length_tmp;
  int header_size;
  int len;
  unsigned char * old;

  header_size = PathCheckHeader(p, &type_tmp, &length_tmp);
  if (!header_size) return (NULL);

  /*
   * ǡtype_tmp äƤΤϥե졼ΥפǤꡤ
   * long/short ΥפǤϤʤΤǡ
   * p->recv_buffer_start += HEADER_SIZE(type_tmp, length_tmp)
   * Τ褦򤷤ƤϤʤ(type_tmp 㤦HEADER_SIZE()裱
   * long/short ΥפǤʤФʤ)
   * PathCheckHeader() Ǥϡtype_tmp ˤñʤե졼ॿפΤߤǼ
   * ֤äƤ(long/short ե졼ɽӥåȤޤǤʤ)Τǡ
   * PathCheckHeader() ͤǤϡlong/short ե졼̤뤳Ȥ
   * ŪˤϤǤʤ
   */
  old = p->recv_buffer_start;
  p->recv_buffer_start += header_size; /* إå򥹥åפ */

  *type = type_tmp;
  *length = length_tmp;

  len = header_size + length_tmp;
  len = ALIGNMENT4(len);

  return (old + len);
}

/*---------------------------------------------------------------------------*/
/* ե졼̵뤹롥(Υե졼˥å)                              */
/*---------------------------------------------------------------------------*/

/* ǡ̵ˤϡ0 ֤ */

int PathDiscardData(Path p, unsigned short int * type)
{
  char * next_start_point;
  unsigned short int tp, len;

  next_start_point = PathReadHeader(p, &tp, &len);
  if (next_start_point == NULL) return (0);

  *type = tp;

  /* ХȤ˥饤Ȥ */
  p->recv_buffer_start = next_start_point;

  return (1);
}

/*---------------------------------------------------------------------------*/
/* ʤǡ(type Τ߼)                                 */
/*---------------------------------------------------------------------------*/

/* ǡ̵ˤϡ0 ֤ */

int PathRecvNoneData(Path p, unsigned short int * type)
{
  char * next_start_point;
  unsigned short int tp, len;

  next_start_point = PathReadHeader(p, &tp, &len);
  if (next_start_point == NULL) return (0);

  *type = tp;

  /* ХȤ˥饤Ȥ */
  p->recv_buffer_start = next_start_point;

  return (1);
}

/*---------------------------------------------------------------------------*/
/* short Υǡ                                                      */
/*---------------------------------------------------------------------------*/

/* ǡ̵ˤϡ0 ֤ */

int PathRecvShortData(Path p, unsigned short int * type,
		      unsigned short int * s1)
{
  char * next_start_point;
  unsigned short int tp, len;

  next_start_point = PathReadHeader(p, &tp, &len);
  if (next_start_point == NULL) return (0);

  *type = tp;
  *s1 = PathGetShortData(p);

  /* ХȤ˥饤Ȥ */
  p->recv_buffer_start = next_start_point;

  return (1);
}

/*---------------------------------------------------------------------------*/
/* short, short Υǡ                                               */
/*---------------------------------------------------------------------------*/

/* ǡ̵ˤϡ0 ֤ */

int PathRecvShortShortData(Path p, unsigned short int * type,
			   unsigned short int * s1,
			   unsigned short int * s2)
{
  char * next_start_point;
  unsigned short int tp, len;

  next_start_point = PathReadHeader(p, &tp, &len);
  if (next_start_point == NULL) return (0);

  *type = tp;
  *s1 = PathGetShortData(p);
  *s2 = PathGetShortData(p);

  /* ХȤ˥饤Ȥ */
  p->recv_buffer_start = next_start_point;

  return (1);
}

/*---------------------------------------------------------------------------*/
/* short, short, short Υǡ                                        */
/*---------------------------------------------------------------------------*/

/* ǡ̵ˤϡ0 ֤ */

int PathRecvShortShortShortData(Path p, unsigned short int * type,
				unsigned short int * s1,
				unsigned short int * s2,
				unsigned short int * s3)
{
  char * next_start_point;
  unsigned short int tp, len;

  next_start_point = PathReadHeader(p, &tp, &len);
  if (next_start_point == NULL) return (0);

  *type = tp;
  *s1 = PathGetShortData(p);
  *s2 = PathGetShortData(p);
  *s3 = PathGetShortData(p);

  /* ХȤ˥饤Ȥ */
  p->recv_buffer_start = next_start_point;

  return (1);
}

/*---------------------------------------------------------------------------*/
/* short, short, short, short Υǡ                                 */
/*---------------------------------------------------------------------------*/

/* ǡ̵ˤϡ0 ֤ */

int PathRecvShortShortShortShortData(Path p, unsigned short int * type,
				     unsigned short int * s1,
				     unsigned short int * s2,
				     unsigned short int * s3,
				     unsigned short int * s4)
{
  char * next_start_point;
  unsigned short int tp, len;

  next_start_point = PathReadHeader(p, &tp, &len);
  if (next_start_point == NULL) return (0);

  *type = tp;
  *s1 = PathGetShortData(p);
  *s2 = PathGetShortData(p);
  *s3 = PathGetShortData(p);
  *s4 = PathGetShortData(p);

  /* ХȤ˥饤Ȥ */
  p->recv_buffer_start = next_start_point;

  return (1);
}

/*---------------------------------------------------------------------------*/
/* short, short, short, short, short Υǡ                          */
/*---------------------------------------------------------------------------*/

/* ǡ̵ˤϡ0 ֤ */

int PathRecvShortShortShortShortShortData(Path p, unsigned short int * type,
					  unsigned short int * s1,
					  unsigned short int * s2,
					  unsigned short int * s3,
					  unsigned short int * s4,
					  unsigned short int * s5)
{
  char * next_start_point;
  unsigned short int tp, len;

  next_start_point = PathReadHeader(p, &tp, &len);
  if (next_start_point == NULL) return (0);

  *type = tp;
  *s1 = PathGetShortData(p);
  *s2 = PathGetShortData(p);
  *s3 = PathGetShortData(p);
  *s4 = PathGetShortData(p);
  *s5 = PathGetShortData(p);

  /* ХȤ˥饤Ȥ */
  p->recv_buffer_start = next_start_point;

  return (1);
}

/*---------------------------------------------------------------------------*/
/* long Υǡ                                                       */
/*---------------------------------------------------------------------------*/

/* ǡ̵ˤϡ0 ֤ */

int PathRecvLongData(Path p, unsigned short int * type,
		     unsigned long int * s1)
{
  char * next_start_point;
  unsigned short int tp, len;

  next_start_point = PathReadHeader(p, &tp, &len);
  if (next_start_point == NULL) return (0);

  *type = tp;
  *s1 = PathGetLongData(p);

  /* ХȤ˥饤Ȥ */
  p->recv_buffer_start = next_start_point;

  return (1);
}

/*---------------------------------------------------------------------------*/
/* long, long Υǡ                                                 */
/*---------------------------------------------------------------------------*/

/* ǡ̵ˤϡ0 ֤ */

int PathRecvLongLongData(Path p, unsigned short int * type,
			 unsigned long int * s1,
			 unsigned long int * s2)
{
  char * next_start_point;
  unsigned short int tp, len;

  next_start_point = PathReadHeader(p, &tp, &len);
  if (next_start_point == NULL) return (0);

  *type = tp;
  *s1 = PathGetLongData(p);
  *s2 = PathGetLongData(p);

  /* ХȤ˥饤Ȥ */
  p->recv_buffer_start = next_start_point;

  return (1);
}

/*---------------------------------------------------------------------------*/
/* char, char, char, char Υǡ                                     */
/*---------------------------------------------------------------------------*/

int PathRecvCharCharCharCharData(Path p, unsigned short int * type,
				 unsigned char * c1, unsigned char * c2,
				 unsigned char * c3, unsigned char * c4)
{
  char * next_start_point;
  unsigned short int tp, len;

  next_start_point = PathReadHeader(p, &tp, &len);
  if (next_start_point == NULL) return (0);

  *type = tp;
  *c1 = PathGetCharData(p);
  *c2 = PathGetCharData(p);
  *c3 = PathGetCharData(p);
  *c4 = PathGetCharData(p);

  /* ХȤ˥饤Ȥ */
  p->recv_buffer_start = next_start_point;

  return (1);
}

/*---------------------------------------------------------------------------*/
/* char, char, short, short Υǡ                                   */
/*---------------------------------------------------------------------------*/

int PathRecvCharCharShortShortData(Path p, unsigned short int * type,
				   unsigned char * c1,
				   unsigned char * c2,
				   unsigned short int * s1,
				   unsigned short int * s2)
{
  char * next_start_point;
  unsigned short int tp, len;

  next_start_point = PathReadHeader(p, &tp, &len);
  if (next_start_point == NULL) return (0);

  *type = tp;
  *c1 = PathGetCharData(p);
  *c2 = PathGetCharData(p);
  *s1 = PathGetShortData(p);
  *s2 = PathGetShortData(p);

  /* ХȤ˥饤Ȥ */
  p->recv_buffer_start = next_start_point;

  return (1);
}

/*---------------------------------------------------------------------------*/
/* char, short, short Υǡ                                         */
/*---------------------------------------------------------------------------*/

int PathRecvCharShortShortData(Path p, unsigned short int * type,
			       unsigned char * c1,
			       unsigned short int * s1,
			       unsigned short int * s2)
{
  /*
   * PATH_BUFFER_ALIGNMENT Ƥˤϡ饤
   * ƤΤǡ
   * PathRecvCharCharShortShortData()Ǽ롥
   */
#ifdef PATH_BUFFER_ALIGNMENT
  unsigned char c2;
  int ret;
  ret = PathRecvCharCharShortShortData(p, type, c1, &c2, s1, s2);
  return (ret);
#else
  char * next_start_point;
  unsigned short int tp, len;

  next_start_point = PathReadHeader(p, &tp, &len);
  if (next_start_point == NULL) return (0);

  *type = tp;
  *c1 = PathGetCharData(p);
  *s1 = PathGetShortData(p);
  *s2 = PathGetShortData(p);

  p->recv_buffer_start = next_start_point;

  return (1);
#endif
}

/*---------------------------------------------------------------------------*/
/* ʸ                                                              */
/*---------------------------------------------------------------------------*/

int PathRecvStringData(Path p, unsigned short int * type,
		       unsigned char * string, int size)
{
  char * next_start_point;
  unsigned short int tp, len;
  unsigned short int str_length;

  next_start_point = PathReadHeader(p, &tp, &len);
  if (next_start_point == NULL) return (0);

  str_length = len;
  if (str_length + 1 > size) ErrorExit("Buffer size is too small.");
  *type = tp;
  memcpy(string, p->recv_buffer_start, str_length);
  string[str_length] = '\0'; /* '\0' ϤĤƤʤΤǡĤɬפ */

  /* ХȤ˥饤Ȥ */
  p->recv_buffer_start = next_start_point;

  return (1);
}

/*---------------------------------------------------------------------------*/
/* short, string Υǡ                                              */
/*---------------------------------------------------------------------------*/

int PathRecvShortStringData(Path p, unsigned short int * type,
			    unsigned short int * s1,
			    unsigned char * string,
			    int size)
{
  char * next_start_point;
  unsigned short int tp, len;
  unsigned short int str_length;

  next_start_point = PathReadHeader(p, &tp, &len);
  if (next_start_point == NULL) return (0);

  str_length = len - (sizeof(*s1));
  if (str_length + 1 > size) ErrorExit("Buffer size is too small.");

  *type = tp;
  *s1 = PathGetShortData(p);
  memcpy(string, p->recv_buffer_start, str_length);
  string[str_length] = '\0'; /* '\0' ϤĤƤʤΤǡĤɬפ */

  /* ХȤ˥饤Ȥ */
  p->recv_buffer_start = next_start_point;

  return (1);
}

/*---------------------------------------------------------------------------*/
/* char, char, short, short, string Υǡ                           */
/*---------------------------------------------------------------------------*/

int PathRecvCharCharShortShortStringData(Path p, unsigned short int * type,
					 unsigned char * c1,
					 unsigned char * c2,
					 unsigned short int * s1,
					 unsigned short int * s2,
					 unsigned char * string,
					 int size)
{
  char * next_start_point;
  unsigned short int tp, len;
  unsigned short int str_length;

  next_start_point = PathReadHeader(p, &tp, &len);
  if (next_start_point == NULL) return (0);

  str_length = len - (sizeof(*c1) + sizeof(*c2) + sizeof(*s1) + sizeof(*s2));
  if (str_length + 1 > size) ErrorExit("Buffer size is too small.");

  *type = tp;
  *c1 = PathGetCharData(p);
  *c2 = PathGetCharData(p);
  *s1 = PathGetShortData(p);
  *s2 = PathGetShortData(p);
  memcpy(string, p->recv_buffer_start, str_length);
  string[str_length] = '\0'; /* '\0' ϤĤƤʤΤǡĤɬפ */

  /* ХȤ˥饤Ȥ */
  p->recv_buffer_start = next_start_point;

  return (1);
}

/*---------------------------------------------------------------------------*/
/* short, short, short, short, string Υǡ                         */
/*---------------------------------------------------------------------------*/

int PathRecvShortShortShortShortStringData(Path p, unsigned short int * type,
					   unsigned short int * s1,
					   unsigned short int * s2,
					   unsigned short int * s3,
					   unsigned short int * s4,
					   unsigned char * string,
					   int size)
{
  char * next_start_point;
  unsigned short int tp, len;
  unsigned short int str_length;

  next_start_point = PathReadHeader(p, &tp, &len);
  if (next_start_point == NULL) return (0);

  str_length = len - (sizeof(*s1) + sizeof(*s2) + sizeof(*s3) + sizeof(*s4));
  if (str_length + 1 > size) ErrorExit("Buffer size is too small.");

  *type = tp;
  *s1 = PathGetShortData(p);
  *s2 = PathGetShortData(p);
  *s3 = PathGetShortData(p);
  *s4 = PathGetShortData(p);
  memcpy(string, p->recv_buffer_start, str_length);
  string[str_length] = '\0'; /* '\0' ϤĤƤʤΤǡĤɬפ */

  /* ХȤ˥饤Ȥ */
  p->recv_buffer_start = next_start_point;

  return (1);
}

/*****************************************************************************/
/* ޤ                                                                  */
/*****************************************************************************/

/*****************************************************************************/
/* End of File.                                                              */
/*****************************************************************************/
