/*****************************************************************************/
/* Socket.c - A library for network communication.                           */
/*                                                                           */
/* Socket.c Copyright (c) 2000-2001 Sakai Hiroaki.                           */
/* All Rights Reserved.                                                      */
/*****************************************************************************/

/*****************************************************************************/
/* Socket 饹(å̿ѤΥ饹)Υ饤֥                   */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                   */
/*****************************************************************************/

#include "SocketP.h"
#include "etc.h"

/*****************************************************************************/
/* ؿ(˸ʤ)                                          */
/*****************************************************************************/

/*===========================================================================*/
/* Socket ֥ȤΥåȤȥɥ쥹롥                     */
/* :0 (顼ˤ 1)                                                 */
/*===========================================================================*/

static int SocketMakeSocketAndAddress(Socket sockt)
{
  sockt->socket = socket(AF_INET, SOCK_STREAM, 0);
  if (sockt->socket == -1) {
    sockt = SocketDestroy(sockt);
    ErrorExit("Cannot create socket.");
  }

  sockt->host = gethostbyname(sockt->hostname);
  if (sockt->host == NULL) {
    sockt = SocketDestroy(sockt);
    ErrorExit("Fail gethostbyname().");
  }

  memset((char *)(&(sockt->address)), 0, sizeof(sockt->address));
  sockt->address.sin_family = AF_INET;
  /* sockt->address.sin_addr = ? See ftpd.c */
  sockt->address.sin_port = htons(sockt->port);
  memmove(&(sockt->address.sin_addr),
	  sockt->host->h_addr,
	  sockt->host->h_length);

  return (0);
}

static int SocketBind(Socket sockt)
{
  if (bind(sockt->socket, (struct sockaddr *)(&(sockt->address)),
	   sizeof(sockt->address)) == -1) {
    sockt = SocketDestroy(sockt);
    ErrorExit("Fail bind().");
  }

  return (0);
}

static int SocketListen(Socket sockt, int backlog)
{
  if (listen(sockt->socket, backlog) == -1) {
    sockt = SocketDestroy(sockt);
    ErrorExit("Fail listen().");
  }

  return (0);
}

static int SocketConnect(Socket sockt)
{
  if (connect(sockt->socket, (struct sockaddr *)(&(sockt->address)),
	      sizeof(sockt->address)) == -1) {
    sockt = SocketDestroy(sockt);
    ErrorExit("Fail connect().");
  }

  return (0);
}

/*===========================================================================*/
/* åȤΥե饰                                                      */
/*===========================================================================*/

/*---------------------------------------------------------------------------*/
/* TCP_NODELAY ե饰                                                  */
/*---------------------------------------------------------------------------*/

int SocketSetTcpNodelay(Socket sockt, int flag)
{
  struct protoent *pe;

#if 1
  pe = getprotobyname("TCP");
  if (pe == NULL) {
    ErrorExit("Cannot get protoname.");
  }
  if (setsockopt(sockt->socket,
#if 1
                 IPPROTO_TCP,
#else
                 pe->p_proto,
#endif
                 TCP_NODELAY, &flag, sizeof(flag)) < 0) {
    ErrorExit("Cannot set socket option.");
  }
#endif

  return (0);
}

/*****************************************************************************/
/* ؿ(˸)                                            */
/*****************************************************************************/

/*===========================================================================*/
/* åֹ֤                                                      */
/*===========================================================================*/

int SocketGetSocket(Socket this)
{
  return (this->socket);
}

/*===========================================================================*/
/* Socket ֥Ȥ롥                                         */
/*===========================================================================*/

static Socket SocketNew()
{
  Socket sockt;

  sockt = (Socket)malloc(sizeof(_Socket));
  if (sockt == NULL) ErrorExit("Cannot allocate memory for socket.");

  sockt->port = 0;
  sockt->socket = 0;
  sockt->hostname = NULL;
  sockt->host = NULL;

  return (sockt);
}

Socket SocketCreate(char * hostname, unsigned short int port)
{
  Socket sockt;
  char buf[200];

  sockt = SocketNew();
  sockt->port = port;

  /* ۥ̾μ */
  if (!hostname) {
    gethostname(buf, 200);
    hostname = buf;
  }
  sockt->hostname = strdup(hostname);
  if (sockt->hostname == NULL) {
    sockt = SocketDestroy(sockt);
    ErrorExit("Cannot copy hostname.");
  }

  if (SocketMakeSocketAndAddress(sockt)) {
    sockt = SocketDestroy(sockt);
    ErrorExit("Cannot make socket.");
  }

  return (sockt);
}

Socket SocketCreateToAccept(char * hostname, unsigned short int port,
			    int backlog)
{
  Socket s;

  s = SocketCreate(hostname, port);
  if (!s) {
    s = SocketDestroy(s);
    ErrorExit("Cannot create socket.");
  }

  if (SocketBind(s)) {
    s = SocketDestroy(s);
    ErrorExit("Cannot bind.");
  }

  if (SocketListen(s, backlog)) {
    s = SocketDestroy(s);
    ErrorExit("Cannot listen.");
  }

  return (s);
}

Socket SocketCreateToConnect(char * hostname, unsigned short int port)
{
  Socket s;

  s = SocketCreate(hostname, port);
  if (!s) {
    s = SocketDestroy(s);
    ErrorExit("Cannot create socket.");
  }

  /*
   * 饤¦Ǥϡݡֹ bind() ɬפ̵(ƤϤʤ)
   * Τǡ bind() ƤϤʤ
   * ݡֹ connect() ǥƥФ٤
   * (Programming UNIX Sockets in C - Frequently Asked Questions ܸ
   *   3.6 )
   */

  if (SocketConnect(s)) {
    s = SocketDestroy(s);
    ErrorExit("Cannot connect.");
  }

  return (s);
}

/*===========================================================================*/
/* accept() μ¹ԡ                                                         */
/*===========================================================================*/

Socket SocketCreateByAccept(Socket sockt)
{
  int len;
  int s;
  Socket sock;

  len = sizeof(sockt->address);
  s = accept(sockt->socket, (struct sockaddr *)(&sockt->address), &len);

  sock = SocketNew();
  sock->socket = s;

  return (sock);
}

/*===========================================================================*/
/* Socket ֥Ȥõ롥                                         */
/*===========================================================================*/

Socket SocketDestroy(Socket sockt)
{
  if (sockt == NULL) return (NULL);

  if (sockt->hostname != NULL) free(sockt->hostname);
  if (sockt->socket > 0) close(sockt->socket); /* ɤ close() 뤫 */

  free(sockt);

  return (NULL);
}

/*****************************************************************************/
/* ޤ                                                                  */
/*****************************************************************************/

/*****************************************************************************/
/* End of File.                                                              */
/*****************************************************************************/
