/*****************************************************************************/
/* N.A.D.A.R. - A network tank battle game.                                  */
/*                                                                           */
/* N.A.D.A.R. Copyright (c) 2000-2001 Sakai Hiroaki.                         */
/* All Rights Reserved.                                                      */
/*                                                                           */
/* N.A.D.A.R. is a network tank battle game. You get into a tank,            */
/* operate it, and destroy other tanks to survive!                           */
/*****************************************************************************/

/*****************************************************************************/
/* Stage 饹(ѤΥ饹)Υ饤֥                        */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                   */
/*****************************************************************************/

#include "StageP.h"

/*****************************************************************************/
/* ؿ                                                                */
/*****************************************************************************/

/*===========================================================================*/
/* ѿμ                                                    */
/*===========================================================================*/

Map StageGetMap(Stage s) {return (s->map);}
ObjList StageGetPlayerList(   Stage s) {return (s->player_list   );}
ObjList StageGetTankList(     Stage s) {return (s->tank_list     );}
ObjList StageGetCanonList(    Stage s) {return (s->canon_list    );}
ObjList StageGetMissileList(  Stage s) {return (s->missile_list  );}
ObjList StageGetExplosionList(Stage s) {return (s->explosion_list);}

/*===========================================================================*/
/* ֥ȤȺ                                                  */
/*===========================================================================*/

/*---------------------------------------------------------------------------*/
/* ֥Ȥ                                                        */
/*---------------------------------------------------------------------------*/

int StageCreateTankForPlayer(Stage stage, Player player)
{
  Tank tank;
  int x, y, tx, ty, f;
  ObjListData current;
  double rx, ry, dx, dy, r;
  Tank other;

  r = ((double)TANK_SIZE / CHARACTER_SIZE) * 2.0;

  /* ᤯̵̤֤ɤå */
  do {
    rx = GetRandomValue(stage->map_width ) + 0.5;
    ry = GetRandomValue(stage->map_height) + 0.5;
    f = 0;
    for (current = ObjList_GetStart(stage->tank_list);
	 !ObjList_IsEndEdge(stage->tank_list, current);
	 current = ObjListData_GetNext(current)) {
      other = (Tank)ObjListData_GetObj(current);
      dx = TankGetX(other) - rx;
      dy = TankGetY(other) - ry;
      if (dx * dx + dy * dy <= r * r) {
	f = 1;
	break;
      }
    }
  } while (f);

  tank = TankCreate(rx, ry, GetRandomValue(TANK_ANGLE_NUMBER),
		    player, TANK_ACTIVE, stage->backward);
  if (tank == NULL) ErrorExit("Cannot create tank.");
  PlayerSetTank(player, tank);
  ObjList_InsertObjToEnd(stage->tank_list, tank, (ObjDestructor)TankDestroy);

  /* ֤μ򥯥ꥢ */
  for (x = -1; x <= 1; x++) {
    for (y = -1; y <= 1; y++) {
      tx = ConvertXFromPositionToMap(TankGetX(tank)) + x;
      ty = ConvertYFromPositionToMap(TankGetY(tank)) + y;
      if (MapIsInside(stage->map, tx, ty))
	MapChangeState(stage->map, tx, ty, 0);
    }
  }
  return (0);
}

Stage StageCreate(int stage_number, ObjList player_list,
		  int map_width, int map_height, int width, int height,
		  int peeping, int backward,
		  int interval_time, int timeout_time, int max_players)
{
  Stage stage;
  Player player;
  ObjListData current;

  stage = (Stage)malloc(sizeof(_Stage));
  if (stage == NULL) return (NULL);

  stage->map_width   = map_width;
  stage->map_height  = map_height;
  stage->width       = width;
  stage->height      = height;
  stage->peeping     = peeping;
  stage->backward    = backward;
  stage->interval    = interval_time;
  stage->timeout     = timeout_time;
  stage->max_players = max_players;

  stage->map = MapCreate(stage, stage->map_width, stage->map_height,
			 stage_number);
  if (stage->map == NULL) ErrorExit("Cannot create stage->map.");

  stage->stage_number = stage_number;
  stage->player_list = player_list;
  stage->tank_list      = ObjList_Create();
  stage->canon_list     = ObjList_Create();
  stage->missile_list   = ObjList_Create();
  stage->explosion_list = ObjList_Create();

  for (current = ObjList_GetStart(player_list);
       !ObjList_IsEndEdge(player_list, current);
       current = ObjListData_GetNext(current)) {
    player = (Player)ObjListData_GetObj(current);
    if (PlayerGetState(player) != PLAYER_ACTIVE) continue;
    StageCreateTankForPlayer(stage, player);
  }

  return (stage);
}

/*---------------------------------------------------------------------------*/
/* ֥Ȥκ                                                        */
/*---------------------------------------------------------------------------*/

Stage StageDestroy(Stage stage)
{
  ObjListData current;
  Player player;

  if (stage == NULL) return (NULL);

  if (stage->map) MapDestroy(stage->map);

  for (current = ObjList_GetStart(stage->player_list);
       !ObjList_IsEndEdge(stage->player_list, current);
       current = ObjListData_GetNext(current)) {
    player = (Player)ObjListData_GetObj(current);
    PlayerSetTank(player, NULL);
  }
  if (stage->tank_list)      ObjList_Destroy(stage->tank_list);
  if (stage->canon_list)     ObjList_Destroy(stage->canon_list);
  if (stage->missile_list)   ObjList_Destroy(stage->missile_list);
  if (stage->explosion_list) ObjList_Destroy(stage->explosion_list);

  free(stage);

  return (NULL);
}

/*---------------------------------------------------------------------------*/
/* Stage is over?                                                            */
/*---------------------------------------------------------------------------*/

static int StageIsOver(Stage stage, int lose)
{
  int n;

  n = ObjList_GetLength(stage->tank_list);

  switch (n) {
  case 0 :
    usleep(100000);
    return (0);
  case 1 :
    return (lose >= 1);
  case 2 :
    return (lose >= 1);
  case 3 :
    return (lose >= 2);
  case 4 :
    return (lose >= 2);
  case 5 :
    return (lose >= 3);
  default :
    return (lose >= n * 2 / 3);
  }
  return (0);
}

/*---------------------------------------------------------------------------*/
/* ƥȤɽ                                                            */
/*---------------------------------------------------------------------------*/

static int StageSendTextData(Stage stage, Path path)
{
  char buffer[PLAYER_NAME_MAX_LENGTH + 100];
  Player player;
  ObjListData current;

  PathSendNoneData(path, TEXT_CLEAR);
  PathSendShortStringData(path, TEXT_DRAW, 0,
			  "      N.A.D.A.R.\n  Copyright (c)\n");
  PathSendShortStringData(path, TEXT_DRAW, 0,
			  "           2000-2001\n    Sakai Hiroaki.\n\n");

  sprintf(buffer, "   Stage : %d\n\n", stage->stage_number + 1);
  PathSendShortStringData(path, TEXT_DRAW, 0, buffer);

  for (current = ObjList_GetStart(stage->player_list);
       !ObjList_IsEndEdge(stage->player_list, current);
       current = ObjListData_GetNext(current)) {
    player = (Player)ObjListData_GetObj(current);
    if (PlayerGetState(player) != PLAYER_ACTIVE) continue;
    if (strlen(PlayerGetName(player)) > PLAYER_NAME_MAX_LENGTH) {
      ErrorExit("Player\'s name is too long.");
    }
    sprintf(buffer, "   %s\n   Win:%d Lose:%d\n\n",
	    PlayerGetName(player),
	    PlayerGetWin(player),
	    PlayerGetLose(player));
    PathSendShortStringData(path, TEXT_DRAW, PlayerGetColorNumber(player),
			    buffer);
  }

  return (0);
}

/*===========================================================================*/
/* ơγ(ФΥᥤ롼)                                      */
/*===========================================================================*/

#define BUFFER_SIZE (PLAYER_NAME_MAX_LENGTH + 32)

int StageStartAsServer(Stage stage, ObjList player_list, Path request)
{
  unsigned short int command;
  unsigned short int length;
  int ret;
  int state;
  Path path;
  Player player;
  Player player2;
  Tank tank;
  Canon canon;
  Missile missile;
  Explosion explosion;
  int lose = 0;
  int f0, f1, f2;
  ObjListData current, current2;
  unsigned long int l1;
  long int interval_timer;
  long int timeout_timer;
  unsigned long int now_time;
  char player_name[BUFFER_SIZE];

  /*
   * ХåեΥꥢ
   * Хåե򥯥ꥢƤʤȡơ׵᤬ĤäƤޤ
   * Ѥ߽ŤʤäƤäƤޤǽ롥
   * (SEND_KEY Ф NOW_KEY Υ꥿󤬤ޤäƤޤ̥ꥢ
   *  ӤȿߤʤäƤޤäƤ)
   */

  for (current = ObjList_GetStart(player_list);
       !ObjList_IsEndEdge(player_list, current);
       current = ObjListData_GetNext(current)) {
    player = (Player)ObjListData_GetObj(current);
    if (PlayerGetState(player) != PLAYER_ACTIVE) continue;

    while (PathCheckData(PlayerGetPath(player), &command, &length)) {

      switch (command) {

      case DELETE_ME :

	PathRecvNoneData(PlayerGetPath(player), &command);
	PathSendNoneData(PlayerGetPath(player), DELETE_YOU);
	PathFlushSendBuffer(PlayerGetPath(player));
	if (TankGetState(PlayerGetTank(player)) == TANK_ACTIVE)
	  TankSetState(PlayerGetTank(player), TANK_EXPLOSION);
	PlayerSetState(player, PLAYER_DESTROY);
	break;

      default :
	/*
	 * ХåեˤޤäƤ׵򥯥ꥢ롥
	 * (ơ׵᤬ĤäƤޤɤɤѤ߽ŤʤäƤȤ)
	 */
	PathDiscardData(PlayerGetPath(player), &command);
	break;
      }
    }
  }

  for (current = ObjList_GetStart(player_list);
       !ObjList_IsEndEdge(player_list, current);
       current = ObjListData_GetNext(current)) {
    player = (Player)ObjListData_GetObj(current);
    if (PlayerGetState(player) != PLAYER_ACTIVE) continue;

    /* ǥץ쥤Υꥢ */
    PathSendNoneData(PlayerGetPath(player), DISPLAY_CLEAR);

    /* ɽƥȥǡ */
    StageSendTextData(stage, PlayerGetPath(player));
  }

  /* ޥåץǡ */
  MapSendChangesToAllPlayers(stage->map, stage->player_list);

  for (current = ObjList_GetStart(player_list);
       !ObjList_IsEndEdge(player_list, current);
       current = ObjListData_GetNext(current)) {
    player = (Player)ObjListData_GetObj(current);
    if (PlayerGetState(player) != PLAYER_ACTIVE) continue;
#if 0 /* ̺̿︺ΤᡤSEND_KEY  XSync() ԤREAD_KEY פˤ */
    PathSendNoneData(PlayerGetPath(player), READ_KEY);
#endif
    PathSendNoneData(PlayerGetPath(player), SEND_KEY);

    /* Хåեեå夷ºݤ˥ǡ */
    PathFlushSendBuffer(PlayerGetPath(player));
  }

  /* ޤν */
  if (stage->interval) stage->interval_timer = GetTimeForMilliSeconds();
  /* if (stage->timeout ) stage->timeout_timer  = GetTimeForMilliSeconds(); */

  while (1) { /* ơΥᥤ롼 */

    /* 饤Ȥμդ */

    if (PathCheckRecvState(request)) {

      path = PathCreateToCommunicateWithClient(request,
					       SERVER_SEND_BUFFER_SIZE,
					       SERVER_RECV_BUFFER_SIZE);

      while (!PathCheckRecvState(path)) { usleep(10000); }
      ret = PathCheckData(path, &command, &length);

      if (ret && (command == JOIN_REQUEST)) {

	PathRecvStringData(path, &command, player_name, BUFFER_SIZE);

	if (ObjList_GetLength(player_list) >= stage->max_players) {
	  PathSendNoneData(path, MANY_PLAYERS);
	  PathFlushSendBuffer(path);
	  path = PathDestroy(path);
	} else {
	  player = PlayerCreate(player_name, path, PLAYER_WAIT);
	  ObjList_InsertObjToEnd(player_list, player,
				 (ObjDestructor)PlayerDestroy);
	  PathSendShortShortShortShortShortData(PlayerGetPath(player),
						JOIN_ACCEPT,
						stage->map_width,
						stage->map_height,
						stage->width,
						stage->height,
						stage->peeping);
	  PathFlushSendBuffer(PlayerGetPath(player));
	}
      } else {
        path = PathDestroy(path);
      }
    }

    /* ȯν */
    for (current = ObjList_GetStart(stage->explosion_list);
	 !ObjList_IsEndEdge(stage->explosion_list, current);
	 current = ObjListData_GetNext(current)) {
      explosion = (Explosion)ObjListData_GetObj(current);
      state = ExplosionAct(explosion, stage->map);
      switch (state) {
      case EXPLOSION_DESTROY :
	current = ObjList_DeleteObjToPrev(stage->explosion_list, current);
	break;
      default :
	break;
      }
    }

    /* ˤƤν */
    for (current = ObjList_GetStart(stage->canon_list);
	 !ObjList_IsEndEdge(stage->canon_list, current);
	 current = ObjListData_GetNext(current)) {
      canon = (Canon)ObjListData_GetObj(current);
      state = CanonAct(canon, stage);
      switch (state) {
      case CANON_EXPLOSION :
	explosion = ExplosionCreate(CanonGetX(canon), CanonGetY(canon),
				    CanonGetTank(canon), EXPLOSION_ACTIVE);
	ObjList_InsertObjToEnd(stage->explosion_list, explosion,
			       (ObjDestructor)ExplosionDestroy);
	CanonSetState(canon, CANON_DESTROY);
	/* Fall through to destroy the object */
      case CANON_DESTROY :
	current = ObjList_DeleteObjToPrev(stage->canon_list, current);
	break;
      default :
	break;
      }
    }

    /* ߥν */
    for (current = ObjList_GetStart(stage->missile_list);
	 !ObjList_IsEndEdge(stage->missile_list, current);
	 current = ObjListData_GetNext(current)) {
      missile = (Missile)ObjListData_GetObj(current);
      state = MissileAct(missile, stage);
      switch (state) {
      case MISSILE_EXPLOSION :
	explosion = ExplosionCreate(MissileGetX(missile),
				    MissileGetY(missile),
				    MissileGetTank(missile),
				    EXPLOSION_ACTIVE);
	ObjList_InsertObjToEnd(stage->explosion_list, explosion,
			       (ObjDestructor)ExplosionDestroy);
	MissileSetState(missile, MISSILE_DESTROY);
	/* Fall through to destroy the object */
      case MISSILE_DESTROY :
	current = ObjList_DeleteObjToPrev(stage->missile_list, current);
	break;
      default :
	break;
      }
    }

    /*
     * Wait ֤Υץ쥤䡼ν
     */
    for (current = ObjList_GetStart(stage->player_list);
	 !ObjList_IsEndEdge(stage->player_list, current);
	 current = ObjListData_GetNext(current)) {

      player = (Player)ObjListData_GetObj(current);
      if (PlayerGetState(player) != PLAYER_WAIT) continue;

      /*
       * Wait ֤ˤʤäƤץ쥤䡼ϾʤΤǡWait ֤ˤʤäƤ
       * ץ쥤䡼ߤĤäȤ GetNowTime() ǻĴ٤нʬ
       * 뤫顤GetNowTime() Ϥ֤(for()֤ɬפ̵)
       */
      now_time = GetNowTime();

      /* 120äĶƤADD_MEƤʤˤϡץ쥤䡼 */
      if (now_time - PlayerGetJoinTime(player) > 120) {
	PathSendNoneData(PlayerGetPath(player), TIMEOUT_YOU);
	PathFlushSendBuffer(PlayerGetPath(player));
	PlayerSetState(player, PLAYER_DESTROY);
	continue;
      }

      if (!PathCheckData(PlayerGetPath(player), &command, &length))
	continue;

      switch (command) {

      case ADD_ME :
	PathRecvNoneData(PlayerGetPath(player), &command);

	/*
	 * MapSendAllMap() ñ˥ޥåפ򤽤ΤޤʤΤǡ
	 * MapChange 塼ޤäƤȡ MapChange äȤ
	 * 򵯤ΤᡤMapSendAllMap ¹Ԥ
	 * MapSendChangesToAllPlayers() ԤMapChange 塼եå
	 * 롥 MapChange ǡϿץ쥤䡼ˤƤϤʤ
	 * ΤǡPLAYER_ACTIVE ˤ˹Ԥɬפ롥
	 * StageCreateTankForPlayer() ǥޥåפѹ뤳ȤΤǡ
	 * MapSendChangesToAllPlayers() ǸƤӽФɬפ롥
	 */
	StageCreateTankForPlayer(stage, player);
	MapSendChangesToAllPlayers(stage->map, stage->player_list);

	PlayerSetState(player, PLAYER_ACTIVE);

	PathSendNoneData(PlayerGetPath(player), DISPLAY_CLEAR);

	MapSendAllMap(stage->map, player);

	/* ɽƥȥǡץ쥤䡼 */
	for (current2 = ObjList_GetStart(stage->player_list);
	     !ObjList_IsEndEdge(stage->player_list, current2);
	     current2 = ObjListData_GetNext(current2)) {
	  player2 = (Player)ObjListData_GetObj(current2);
	  if (PlayerGetState(player2) != PLAYER_ACTIVE) continue;
	  StageSendTextData(stage, PlayerGetPath(player2));
	  PathFlushSendBuffer(PlayerGetPath(player2));
	}

#if 0 /* ̺̿︺ΤᡤSEND_KEY  XSync() ԤREAD_KEY פˤ */
	PathSendNoneData(PlayerGetPath(player), READ_KEY);
#endif
	PathSendNoneData(PlayerGetPath(player), SEND_KEY);

	/* Хåեեå夷ºݤ˥ǡ */
	PathFlushSendBuffer(PlayerGetPath(player));

	break;

      case DELETE_ME :

	PathRecvNoneData(PlayerGetPath(player), &command);
	PathSendNoneData(PlayerGetPath(player), DELETE_YOU);
	PathFlushSendBuffer(PlayerGetPath(player));
	PlayerSetState(player, PLAYER_DESTROY);
	break;

      default :
	ErrorExit("Invalid command.");
      }
    }

    /*
     *  Player ǡȤ롥ǡä Player ϡ
     * State Ū Done ˤƤ
     */
    f1 = 1; /* ץ쥤䡼μ̵ä顤usleep()ԤĤΥե饰 */
    f2 = 1; /* ॢȤνΤΥե饰 */
    do {
      f0 = 0; /* ٤ƤΥץ쥤䡼νѤɤΥե饰 */
      if (!f1) usleep(1000); /* Wait in 1 millisecond. */
      f1 = 0;

      /*
       * νäץ쥤䡼 state  PLAYER_DONE ѹơ
       * ꥹȤƬ˰ư롥
       * Ѥץ쥤䡼ϡꥹȤƬ˻äƤƤΤǡ
       * PLAYER_DONE Ĥä顤ʾ帡ɬפ̵Ȥ
       * Ȥˤʤ롥ΤᡤꥹȤνü鸡롥
       */
      for (current = ObjList_GetEnd(stage->player_list);
	   !ObjList_IsStartEdge(stage->player_list, current);
	   current = ObjListData_GetPrev(current)) {

	player = (Player)ObjListData_GetObj(current);

	  /*
	   * Ѥץ쥤䡼ϡꥹȤƬ˻äƤƤΤǡ
	   * PLAYER_DONE Ĥä顤ʾ帡ɬפ̵Ȥ
	   * ȤʤΤǡbreak 롥
	   */
	if (PlayerGetState(player) == PLAYER_DONE) break;

	if (PlayerGetState(player) != PLAYER_ACTIVE) continue;

	if (!PathCheckData(PlayerGetPath(player), &command, &length)) {
	  f0 = 1;
	  continue;
	}
	f1 = 1;

	switch (command) {

	case DELETE_ME :

	  PathRecvNoneData(PlayerGetPath(player), &command);
	  PathSendNoneData(PlayerGetPath(player), DELETE_YOU);
	  PathFlushSendBuffer(PlayerGetPath(player));
	  if (TankGetState(PlayerGetTank(player)) == TANK_ACTIVE)
	    TankSetState(PlayerGetTank(player), TANK_EXPLOSION);
	  PlayerSetState(player, PLAYER_DESTROY);
	  break;

	case NOW_KEY :

	  PathRecvLongData(PlayerGetPath(player), &command, &l1);
	  TankSetAction(PlayerGetTank(player), l1);
	  PlayerSetState(player, PLAYER_DONE);

	  /*
	   * Ѥץ쥤䡼ϡꥹȤƬ˻äƤ
	   * ˤꡤʲ롥
	   * PLAYER_ACTIVE ˤʤäƤץ쥤䡼򸡺Τˡ
	   *   ꥹȤ򤹤٤Ƹɬפ̵ʤ롥
	   *   (PLAYER_DONE ˤʤäƤȤޤǸФ褤)
	   * ֿ˻֤ץ쥤䡼ꥹȤƬնˤ뤿ᡤ
	   *   ǡ̤ͥˤʤ롥Ĥޤꡤͥåȥ
	   *   ٤ץ쥤䡼ˤ˥ǡ褦ˤʤ롥
	   *   (餬ŪǤ)
	   */
	  current2 = ObjListData_GetNext(current);
	  ObjList_MoveObjToStart(stage->player_list, current);
	  current = current2;
	  break;

	default :
	  ErrorExit("unknown command.");
	}
      }

      /* ΥॢȽ */
      /* ޤǡƤʤץ쥤䡼Ф */
      if (stage->timeout && f0) {
	if (f2) {
	  /* ॢȤΥޤν */
	  stage->timeout_timer = GetTimeForMilliSeconds();
	  f2 = 0;
	} else {
	  timeout_timer = GetTimeForMilliSeconds();
	  if (stage->timeout_timer > timeout_timer)
	    stage->timeout_timer -= 256000;
	  if (timeout_timer - stage->timeout_timer >= stage->timeout)
	    f0 = 0;
	  /* stage->timeout_timer = timeout_timer; */
	}
      }
    } while (f0);

    /* Player  State  Done  Active ᤹ */
    for (current = ObjList_GetStart(stage->player_list);
	 !ObjList_IsEndEdge(stage->player_list, current);
	 current = ObjListData_GetNext(current)) {
      player = (Player)ObjListData_GetObj(current);
      switch (PlayerGetState(player)) {
      case PLAYER_WAIT :
	break;
      case PLAYER_ACTIVE :
	/*
	 *  ACTIVE ȤȤϡॢȤȤȤʤΤǡ
	 * ץ쥤䡼롥
	 */
	PathSendNoneData(PlayerGetPath(player), TIMEOUT_YOU);
	PathFlushSendBuffer(PlayerGetPath(player));
	if (TankGetState(PlayerGetTank(player)) == TANK_ACTIVE)
	  TankSetState(PlayerGetTank(player), TANK_EXPLOSION);
	PlayerSetState(player, PLAYER_DESTROY);
	break;
      case PLAYER_DONE :
	PlayerSetState(player, PLAYER_ACTIVE);
	break;
      default :
	break;
      }
    }

    /* ֤ν */
    for (current = ObjList_GetStart(stage->tank_list);
	 !ObjList_IsEndEdge(stage->tank_list, current);
	 current = ObjListData_GetNext(current)) {
      tank = (Tank)ObjListData_GetObj(current);
      state = TankAct(tank, stage);
      switch (state) {
      case TANK_EXPLOSION :
	explosion = ExplosionCreate(TankGetX(tank), TankGetY(tank),
				    tank, EXPLOSION_ACTIVE);
	ObjList_InsertObjToEnd(stage->explosion_list, explosion,
			       (ObjDestructor)ExplosionDestroy);
	TankSetState(tank, TANK_EXPLOSING);
	TankSetDestroyTimer(tank, TANK_DESTROY_INTERVAL);
	break;
      case TANK_EXPLOSING :
	TankDecDestroyTimer(tank);
	if (TankGetDestroyTimer(tank) <= 0) {
	  TankSetState(tank, TANK_DESTROY);
	  lose++;
	}
	break;
      case TANK_DESTROY :
	/* current = ObjList_DeleteObjToPrev(stage->tank_list, current); */
	break;
      default :
	break;
      }
    }

    /*
     * ơλȽ
     * (SEND_KEY ׵Ф NOW_KEY ֿƤ饹ơλ
     *  ʤȡNOW_KEY ХåեѤ߽ŤʤäƤơʤ
     *  ȿߤʤȤݤƤޤΤǡNOW_KEY ν
     *  Ȥ˥ơλȽƤ)
     */
    if (StageIsOver(stage, lose)) break;

    /*
     * ǡ
     */

    MapSendChangesToAllPlayers(stage->map, stage->player_list);

    for (current = ObjList_GetStart(stage->player_list);
	 !ObjList_IsEndEdge(stage->player_list, current);
	 current = ObjListData_GetNext(current)) {

      player = (Player)ObjListData_GetObj(current);
      if (PlayerGetState(player) != PLAYER_ACTIVE) continue;

      PathSendNoneData(PlayerGetPath(player), GAME_CLEAN);

      for (current2 = ObjList_GetStart(stage->tank_list);
	   !ObjList_IsEndEdge(stage->tank_list, current2);
	   current2 = ObjListData_GetNext(current2)) {
	tank = (Tank)ObjListData_GetObj(current2);
	/* if (TankGetState(canon) != TANK_ACTIVE) continue; */ /* ס */
	PathSendCharCharShortShortData(PlayerGetPath(player), TANK_DRAW,
				       (unsigned char)TankGetAngle(tank),
				       (unsigned char)TankGetColorNumber(tank),
				       TankGetX(tank) * 100 + 0.5,
				       TankGetY(tank) * 100 + 0.5);
      }

      for (current2 = ObjList_GetStart(stage->canon_list);
	   !ObjList_IsEndEdge(stage->canon_list, current2);
	   current2 = ObjListData_GetNext(current2)) {
	canon = (Canon)ObjListData_GetObj(current2);
	if (CanonGetState(canon) != CANON_ACTIVE) continue;
	PathSendShortShortData(PlayerGetPath(player), CANON_DRAW,
			       CanonGetX(canon) * 100 + 0.5,
			       CanonGetY(canon) * 100 + 0.5);
      }

      for (current2 = ObjList_GetStart(stage->missile_list);
	   !ObjList_IsEndEdge(stage->missile_list, current2);
	   current2 = ObjListData_GetNext(current2)) {
	missile = (Missile)ObjListData_GetObj(current2);
	if (MissileGetState(missile) != MISSILE_ACTIVE) continue;
	PathSendShortShortData(PlayerGetPath(player), MISSILE_DRAW,
			       MissileGetX(missile) * 100 + 0.5,
			       MissileGetY(missile) * 100 + 0.5);
      }

      for (current2 = ObjList_GetStart(stage->explosion_list);
	   !ObjList_IsEndEdge(stage->explosion_list, current2);
	   current2 = ObjListData_GetNext(current2)) {
	explosion = (Explosion)ObjListData_GetObj(current2);
	if (ExplosionGetState(explosion) != EXPLOSION_ACTIVE) continue;
	PathSendCharShortShortData(PlayerGetPath(player), EXPLOSION_DRAW,
				   ExplosionGetRadius(explosion),
				   ExplosionGetX(explosion) * 100 + 0.5,
				   ExplosionGetY(explosion) * 100 + 0.5);
      }

      PathSendShortShortData(PlayerGetPath(player), DISPLAY_COPY,
			     TankGetX(PlayerGetTank(player)) * 100 + 0.5,
			     TankGetY(PlayerGetTank(player)) * 100 + 0.5);

#if 0 /* ̺̿︺ΤᡤSEND_KEY  XSync() ԤREAD_KEY פˤ */
      PathSendNoneData(PlayerGetPath(player), READ_KEY);
#endif
      PathSendNoneData(PlayerGetPath(player), SEND_KEY);

      PathFlushSendBuffer(PlayerGetPath(player));
    }

    /* interval 򸫤ơֹ碌 */
    if (stage->interval) {
      while (1) {
	interval_timer = GetTimeForMilliSeconds();
	if (stage->interval_timer > interval_timer)
	  stage->interval_timer -= 256000;
	if (interval_timer - stage->interval_timer >= stage->interval)
	  break;
	usleep((interval_timer - stage->interval_timer) * 100);
      }
      stage->interval_timer = interval_timer;
    }

    usleep(10000);
  }

  return (0);
}

/*****************************************************************************/
/* ޤ                                                                  */
/*****************************************************************************/

/*****************************************************************************/
/* End of File.                                                              */
/*****************************************************************************/
