/*****************************************************************************/
/* N.A.D.A.R. - A network tank battle game.                                  */
/*                                                                           */
/* N.A.D.A.R. Copyright (c) 2000-2001 Sakai Hiroaki.                         */
/* All Rights Reserved.                                                      */
/*                                                                           */
/* N.A.D.A.R. is a network tank battle game. You get into a tank,            */
/* operate it, and destroy other tanks to survive!                           */
/*****************************************************************************/

/*****************************************************************************/
/* Tank 饹(֤ѤΥ饹)Υ饤֥                       */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                   */
/*****************************************************************************/

#include "TankP.h"

/*****************************************************************************/
/* ؿ                                                                */
/*****************************************************************************/

/*===========================================================================*/
/* ѿμ                                                    */
/*===========================================================================*/

double TankGetX(Tank tank)  { return (tank->x); }
double TankGetY(Tank tank)  { return (tank->y); }
int TankGetAngle(Tank tank) { return (tank->angle); }

int TankGetState(Tank tank)              { return (tank->state); }
int TankSetState(Tank tank, int state)   { return (tank->state = state); }
unsigned long int TankGetAction(Tank tank)             { return (tank->action); }
unsigned long int TankSetAction(Tank tank, unsigned long int action) { return (tank->action = action); }

int TankGetColorNumber(Tank tank) { return (tank->color_number); }

Player TankGetPlayer(Tank tank) { return (tank->player); }

int TankGetCanonNumber(Tank tank) { return (tank->canon_number); }
int TankGetCanonActiveNumber(Tank tank) { return (tank->canon_active_number); }
int TankGetMissileNumber(Tank tank) { return (tank->missile_number); }
int TankGetMissileActiveNumber(Tank tank)
{ return (tank->missile_active_number); }

int TankGetDestroyTimer(Tank tank)        { return (tank->destroy_timer); }
int TankSetDestroyTimer(Tank tank, int t) { return (tank->destroy_timer = t); }
int TankDecDestroyTimer(Tank tank)        { return ((tank->destroy_timer)--); }

int TankIncCanon(Tank tank)
{ (tank->canon_number)++; (tank->canon_active_number)++; return (0); }

int TankDecCanon(Tank tank)
{ (tank->canon_active_number)++; return (0); }

int TankIncMissile(Tank tank)
{ (tank->missile_number)++; (tank->missile_active_number)++; return (0); }

int TankDecMissile(Tank tank)
{ (tank->missile_active_number)++; return (0); }

/*===========================================================================*/
/* ȯ                                                                      */
/*===========================================================================*/

static int TankDead(Tank tank, Explosion explosion)
{
  if (explosion == NULL) {
    /* ξ */
    PlayerIncLose(TankGetPlayer(tank));
    PlayerIncLose(TankGetPlayer(tank));
  } else {
    if (TankGetPlayer(ExplosionGetTank(explosion)) == TankGetPlayer(tank)) {
      /* ʬˤƤˤ뼫Ǥξ */
      PlayerIncLose(TankGetPlayer(tank));
      PlayerIncLose(TankGetPlayer(tank));
    } else {
      PlayerIncWin(TankGetPlayer(ExplosionGetTank(explosion)));
      PlayerIncLose(TankGetPlayer(tank));
    }
  }
  tank->state = TANK_EXPLOSION;

  return (0);
}

/*===========================================================================*/
/* ȯȽ                                                                */
/*===========================================================================*/

static int TankCheckExplosion(Tank tank, ObjList explosion_list)
{
  Explosion explosion;
  ObjListData current;

  for (current = ObjList_GetStart(explosion_list);
       !ObjList_IsEndEdge(explosion_list, current);
       current = ObjListData_GetNext(current)) {
    explosion = (Explosion)ObjListData_GetObj(current);
    if (ExplosionCheck(explosion, tank->x, tank->y)) {
      TankDead(tank, explosion);
      return (tank->state);
    }
  }

  return (tank->state);
}

/*===========================================================================*/
/* ֥ȤȺ                                                  */
/*===========================================================================*/

/*---------------------------------------------------------------------------*/
/* ֥Ȥ                                                        */
/*---------------------------------------------------------------------------*/

Tank TankCreate(double x, double y, int angle, Player player, int state,
		int backward)
{
  Tank tank;

  tank = (Tank)malloc(sizeof(_Tank));
  if (tank == NULL) ErrorExit("Cannot create tank.");

  tank->x = x;
  tank->y = y;
  tank->angle = angle;
  tank->state = state;
  tank->action = TANK_NO_ACTION;
  tank->color_number = PlayerGetColorNumber(player);
  tank->player = player;
  tank->canon_number          = 0;
  tank->missile_number        = 0;
  tank->canon_active_number   = 0;
  tank->missile_active_number = 0;
  tank->canon_timer           = 0;
  tank->missile_timer         = 0;
  tank->destroy_timer         = 0;
  tank->backward              = backward;

  return (tank);
}

/*---------------------------------------------------------------------------*/
/* ֥Ȥκ                                                        */
/*---------------------------------------------------------------------------*/

Tank TankDestroy(Tank tank)
{
  if (tank == NULL) return (NULL);

  free(tank);

  return (NULL);
}

/*===========================================================================*/
/* ư                                                                  */
/*===========================================================================*/

/*---------------------------------------------------------------------------*/
/* ٲû                                                                  */
/*---------------------------------------------------------------------------*/

static int TankAddAngle(int angle, int incr)
{
  angle += incr;
  while (angle < 0) angle += TANK_ANGLE_NUMBER;
  angle %= TANK_ANGLE_NUMBER;
  return (angle);
}

/*---------------------------------------------------------------------------*/
/* ư                                                                      */
/*---------------------------------------------------------------------------*/

static int TankMove(Tank tank, Map map, double speed, int direction,
		    ObjList tank_list)
{
  double x, y;
  double c0, s0, c1, s1;
  double x0, y0, x1, y1, x2, y2;
  double dx, dy;
  double size;
  ObjListData current;
  Tank other;

  size = (double)TANK_SIZE / CHARACTER_SIZE;

  x  = tank->x + cos_value[tank->angle] * speed * direction;
  y  = tank->y + sin_value[tank->angle] * speed * direction;
  c0 = cos_value[             tank->angle    ] * (size * 0.5) * direction;
  s0 = sin_value[             tank->angle    ] * (size * 0.5) * direction;
  c1 = cos_value[TankAddAngle(tank->angle, 6)] * (size * 0.5);
  s1 = sin_value[TankAddAngle(tank->angle, 6)] * (size * 0.5);
  x0 = x  + c0; y0 = y  + s0;
  x1 = x0 - c1; y1 = y0 - s1;
  x2 = x0 + c1; y2 = y0 + s1;

  if (MapGetState(map,
		  ConvertXFromPositionToMap(x0),
		  ConvertYFromPositionToMap(y0)) != 0 ||
      MapGetState(map,
		  ConvertXFromPositionToMap(x1),
		  ConvertYFromPositionToMap(y1)) != 0 ||
      MapGetState(map,
		  ConvertXFromPositionToMap(x2),
		  ConvertYFromPositionToMap(y2)) != 0) {

    return (1);
  }

  for (current = ObjList_GetStart(tank_list);
       !ObjList_IsEndEdge(tank_list, current);
       current = ObjListData_GetNext(current)) {
    other = (Tank)ObjListData_GetObj(current);
    if (other == tank) continue;

    dx = other->x - x;
    dy = other->y - y;

    if (dx * dx + dy * dy <= size * size) return (1);
  }

  tank->x = x;
  tank->y = y;
  return (0);
}

/*---------------------------------------------------------------------------*/
/*                                                                       */
/*---------------------------------------------------------------------------*/

static int TankGoForward(Tank tank, Map map, ObjList tank_list)
{
  if (!TankMove(tank, map, TANK_SPEED       , 1, tank_list)) return (0);
  if (!TankMove(tank, map, TANK_SPEED * 0.75, 1, tank_list)) return (0);
  if (!TankMove(tank, map, TANK_SPEED * 0.50, 1, tank_list)) return (0);
  if (!TankMove(tank, map, TANK_SPEED * 0.25, 1, tank_list)) return (0);
  return (1);
}

/*---------------------------------------------------------------------------*/
/*                                                                       */
/*---------------------------------------------------------------------------*/

static int TankGoBackward(Tank tank, Map map, ObjList tank_list)
{
  if (!(tank->backward)) return (1);
  if (!TankMove(tank, map, TANK_SPEED       , -1, tank_list)) return (0);
  if (!TankMove(tank, map, TANK_SPEED * 0.75, -1, tank_list)) return (0);
  if (!TankMove(tank, map, TANK_SPEED * 0.50, -1, tank_list)) return (0);
  if (!TankMove(tank, map, TANK_SPEED * 0.25, -1, tank_list)) return (0);
  return (1);
}

/*---------------------------------------------------------------------------*/
/* ž                                                                    */
/*---------------------------------------------------------------------------*/

static int TankTurnLeft(Tank tank)
{
  tank->angle = TankAddAngle(tank->angle, -1);
  return (tank->angle);
}

/*---------------------------------------------------------------------------*/
/* ž                                                                    */
/*---------------------------------------------------------------------------*/

static int TankTurnRight(Tank tank)
{
  tank->angle = TankAddAngle(tank->angle, 1);
  return (tank->angle);
}

/*---------------------------------------------------------------------------*/
/* ˤȯ                                                                  */
/*---------------------------------------------------------------------------*/

static int TankFireCanon(Tank tank, Stage stage)
{
  Canon canon;

  if (tank->canon_timer) return (1);
  /*
  if (tank->canon_active_number > 4) return (1);
  */

  canon = CanonCreate(tank->x, tank->y,
		      cos_value[tank->angle] * CANON_SPEED,
		      sin_value[tank->angle] * CANON_SPEED,
		      tank, CANON_ACTIVE);

  ObjList_InsertObjToEnd(StageGetCanonList(stage), canon,
			 (ObjDestructor)CanonDestroy);

  CanonAct(canon, stage);

  tank->canon_timer = CANON_INTERVAL;

  return (0);
}

/*---------------------------------------------------------------------------*/
/* ߥȯ                                                              */
/*---------------------------------------------------------------------------*/

static int TankFireMissile(Tank tank, Stage stage)
{
  Missile missile;

  if (tank->missile_timer) return (1);
  if (tank->missile_number > 2) return (1);

  missile = MissileCreate(tank->x, tank->y,
			  cos_value[tank->angle] * MISSILE_SPEED,
			  sin_value[tank->angle] * MISSILE_SPEED,
			  tank, MISSILE_ACTIVE, StageGetTankList(stage));

  ObjList_InsertObjToEnd(StageGetMissileList(stage), missile,
			 (ObjDestructor)MissileDestroy);

  MissileAct(missile, stage);

  tank->missile_timer = MISSILE_INTERVAL;

  return (0);
}

/*---------------------------------------------------------------------------*/
/* ֤ư                                                                */
/*---------------------------------------------------------------------------*/

int TankAct(Tank tank, Stage stage)
{
  switch (tank->state) {
  case TANK_WAIT :
    break;
  case TANK_ACTIVE :
    if (tank->action & TANK_GO_FORWARD  )
      TankGoForward(tank, StageGetMap(stage), StageGetTankList(stage));
    if (tank->action & TANK_GO_BACKWARD )
      TankGoBackward(tank, StageGetMap(stage), StageGetTankList(stage));
    if (tank->action & TANK_TURN_LEFT   ) TankTurnLeft(tank);
    if (tank->action & TANK_TURN_RIGHT  ) TankTurnRight(tank);
    if (tank->action & TANK_FIRE_CANON  ) TankFireCanon(tank, stage);
    if (tank->action & TANK_FIRE_MISSILE) TankFireMissile(tank, stage);
    if (tank->action & TANK_DEAD        ) TankDead(tank, NULL);
    tank->action = TANK_NO_ACTION;
    TankCheckExplosion(tank, StageGetExplosionList(stage));
    break;
  case TANK_EXPLOSION :
    break;
  case TANK_EXPLOSING :
    break;
  case TANK_DESTROY :
    break;
  default :
    break;
  }

  if (tank->canon_timer  ) (tank->canon_timer  )--;
  if (tank->missile_timer) (tank->missile_timer)--;

  return (tank->state);
}

/*****************************************************************************/
/* ޤ                                                                  */
/*****************************************************************************/

/*****************************************************************************/
/* End of File.                                                              */
/*****************************************************************************/
